import unittest
import tifffile as tif
import numpy as np

from pathlib import Path
import matplotlib.pyplot as plt

import vodex as vx
from numan.plots import *

# this folder has the processed folder results for 1v2v3v5 , 7 cycles, 94 volumes each
project_dir = "D:/Code/repos/numan/data/test/ld00_casper/processed"


class TestLabelPlotter(unittest.TestCase):
    experiment = vx.Experiment.load(Path(project_dir, "experiment_raw.db"))

    def test_init(self):
        lp = LabelPlotter(self.experiment, "number")
        names = ["b", "d1", "d2", "d3", "d5"]
        # labels in one cycle
        values = [0, 0, 0, 2,  # 0 - 3
                  0, 0, 0, 0, 0, 4,  # 4- 9
                  0, 0, 0, 0, 0, 0, 1,  # 10 - 16
                  0, 0, 0, 0, 0, 0, 0, 0, 0, 2,  # 17 - 26
                  0, 0, 0, 0, 0, 4,  # 27 - 32
                  0, 0, 0, 0, 0, 0, 3,  # 33 - 39
                  0, 0, 0, 0, 0, 0, 0, 0, 0, 1,
                  0, 0, 0, 0, 0, 3,
                  0, 0, 0, 0, 0, 0, 0, 0, 0, 1,
                  0, 0, 0, 0, 0, 0, 0, 3,
                  0, 0, 0, 0, 0, 0, 0, 0, 0, 4,
                  0, 0, 0, 0, 0, 0, 2,
                  0, 0, 0]

        self.assertListEqual(names, lp.names)
        self.assertListEqual(values, lp.values)

    def test_plot_labels(self):
        lp = LabelPlotter(self.experiment, "number")
        lp.plot_labels(show_plot=True)

    def test_plot_labels_forward_shift(self):
        lp = LabelPlotter(self.experiment, "number")
        lp.plot_labels(forward_shift=1, show_plot=True)

    def test_plot_labels_time_points(self):
        lp = LabelPlotter(self.experiment, "number")
        lp.plot_labels(time_points=[16, 3, 39, 9],
                       show_plot=True)

    def test_plot_labels_forward_shift_time_points(self):
        lp = LabelPlotter(self.experiment, "number")
        lp.plot_labels(forward_shift=1,
                       time_points=[16, 3, 39, 9],
                       show_plot=True)


class TestSignalPlotter(unittest.TestCase):
    experiment = vx.Experiment.load(Path(project_dir, "experiment_raw.db"))
    spots = Spots.from_json(f"{project_dir}/spots/signals/spots_SvB_max.json")

    # initialise the signal plotter with DFF signal
    SLIDING_WINDOW = 15  # in volumes
    signals = spots.get_group_signals(spots.groups["sig2vB"]).as_dff(SLIDING_WINDOW)
    s_plotter = SignalPlotter(signals, experiment, "number")

    def test_prepare_cycle(self):
        trace_id = 0
        trace = self.s_plotter.get_trace(trace_id)
        trace = self.s_plotter.prepare_cycle(trace)
        self.assertTupleEqual((7, 94), trace.shape)

    def test_plot_cycles(self):
        trace_id = 0
        ax_limits = self.s_plotter.plot_cycles(None, trace_id, show_plot=True)
        self.s_plotter.plot_cycles(None, trace_id, show_plot=True, forward_shift=3)
        self.assertAlmostEqual(
            [-0.5, 93.5, -0.004590296974461647, 0.039927868617393523],
            ax_limits)

    def test_prepare_psh(self):
        trace_id = 0
        labels = ["d1", "d2", "d3", "d5"]
        padding = [-2, -1, 0, 1, 2, 3, 4]
        trace = self.s_plotter.get_trace(trace_id)
        trace, label_selection = self.s_plotter.prepare_psh(trace, labels, padding)
        correct_label_selection = [14, 15, 16, 17, 18, 19, 20,
                                   1, 2, 3, 4, 5, 6, 7,
                                   37, 38, 39, 40, 41, 42, 43,
                                   7, 8, 9, 10, 11, 12, 13]
        self.assertListEqual(correct_label_selection, label_selection.tolist())

    def test_plot_psh_b(self):
        trace_id = 0
        labels = ["d1", "d2", "d3", "d5"]
        padding = [-2, -1, 0, 1, 2, 3, 4]
        _ = self.s_plotter.plot_psh(None, trace_id, labels, padding, show_plot=True)

    def test_plot_psh_0(self):
        trace_id = 0
        labels = ["d1", "d2", "d3", "d5"]
        padding = [0]
        _ = self.s_plotter.plot_psh(None, trace_id, labels, padding,
                                    split=False, plot_individual=False, show_plot=True)

    def test_make_psh_figure(self):
        trace_ids = [0, 1, 2, 3, 4]
        labels = ["d1", "d2", "d3", "d5"]
        padding = [-2, -1, 0, 1, 2, 3, 4]
        _ = self.s_plotter.make_psh_figure(trace_ids, labels, padding,
                                           "Main title",
                                           ["0", "1", "2", "3", "4"],
                                           show_plot=True)

    def test_make_cycle_figure(self):
        trace_ids = [0, 1, 2, 3, 4]
        forward_shift = 3
        _ = self.s_plotter.make_cycle_figure(trace_ids,
                                             "Main title",
                                             ["0", "1", "2", "3", "4"],
                                             forward_shift=forward_shift,
                                             show_plot=True)

    def test_make_avg_act_scat_figure(self):
        labels = ["d2", "d3", "d5"]
        _ = self.s_plotter.make_avg_act_scat_figure(labels, "Title",
                                                    figure_layout=[1, 3],
                                                    figsize=(12, 10), dpi=160,
                                                    show_plot=True)
        labels = ["d2", "d3", "d5", "d2"]
        _ = self.s_plotter.make_avg_act_scat_figure(labels, "Title",
                                                    figure_layout=[2, 3],
                                                    figsize=(12, 10), dpi=160,
                                                    show_plot=True)


class TestSignalPlotterV2(unittest.TestCase):
    experiment = vx.Experiment.load(Path(project_dir, "experiment_w_covariates_raw.db"))
    spots = Spots.from_json(f"{project_dir}/spots/signals/spots_SvB_max.json")

    # initialise the signal plotter with DFF signal
    SLIDING_WINDOW = 15  # in volumes
    signals = spots.get_group_signals(spots.groups["sig2vB"]).as_dff(SLIDING_WINDOW)
    s_plotter_v2 = SignalPlotter(signals, experiment, "number")

    def test_get_labels_in_cycle(self):
        labels = ["d1"]
        padding = [0]
        selection = self.s_plotter_v2.get_labels_in_cycle(labels, padding=padding)
        self.assertListEqual(selection.tolist(), [[16], [49], [65]])

        labels = ["d1"]
        padding = [-1, 0, 1]
        selection = self.s_plotter_v2.get_labels_in_cycle(labels,
                                                          padding=padding, shift_by_padding=False)
        self.assertListEqual(selection.tolist(), [[15, 16, 17], [48, 49, 50], [64, 65, 66]])

    def test_prepare_psh(self):
        trace_id = 0
        labels = ["d1", "d2", "d3", "d5"]
        padding = [-2, -1, 0, 1, 2, 3, 4]
        trace = self.s_plotter_v2.get_trace(trace_id)
        trace, label_selection = self.s_plotter_v2.prepare_psh(trace, labels, padding)
        correct_label_selection = [14, 15, 16, 17, 18, 19, 20,
                                   1, 2, 3, 4, 5, 6, 7,
                                   37, 38, 39, 40, 41, 42, 43,
                                   7, 8, 9, 10, 11, 12, 13]
        self.assertListEqual(correct_label_selection, label_selection.tolist())

    def test_plot_psh_b(self):
        trace_id = 0
        labels = ["d1", "d2", "d3", "d5"]
        conditions = [
            [("number", "d1")],
            [("number", "d2")],
            [("number", "d3")],
            [("number", "d5")]
        ]
        padding = [-2, -1, 0, 1, 2, 3, 4]
        _ = self.s_plotter_v2.plot_psh(None, trace_id, labels, padding, show_plot=True)

    def test_plot_psh_0(self):
        trace_id = 0
        labels = ["d1", "d2", "d3", "d5"]
        padding = [0]
        _ = self.s_plotter_v2.plot_psh(None, trace_id, labels, padding,
                                       split=False, plot_individual=False, show_plot=True)

    def test_prepare_condition_trace(self):
        conditions = [
            [("number", "d1")],
            [("number", "d2")],
            [("number", "d3")],
            [("number", "d5")]
        ]
        padding = [-2, -1, 0, 1, 2, 3, 4]
        trace_id = 0
        # get the individual signal trace
        trace = self.s_plotter_v2.get_trace(trace_id)

        condition_signals = self.s_plotter_v2.prepare_condition_trace(trace,
                                                                      conditions,
                                                                      padding)
        self.assertTupleEqual(condition_signals.shape, (21, 28))

    def test_get_labels_for_conditions(self):
        conditions = [
            [("number", "d1")],
            [("number", "d2")],
            [("number", "d3")],
            [("number", "d5")]
        ]
        condition_label = ["d1", "d2", "d3", "d5"]
        cl = self.s_plotter_v2.get_labels_for_conditions(conditions)
        self.assertListEqual(condition_label, cl)

    def test_get_volumes_for_conditions_1c(self):
        conditions = [
            [("number", "d1")],
            [("number", "d2")],
            [("number", "d3")],
            [("number", "d5")]
        ]

        vols = self.s_plotter_v2.get_volumes_for_conditions(conditions)
        correct_vols = [[16, 49, 65, 110, 143, 159, 204, 237, 253, 298, 331,
                         347, 392, 425, 441, 486, 519, 535, 580, 613, 629],
                        [3, 26, 90, 97, 120, 184, 191, 214, 278, 285, 308, 372,
                         379, 402, 466, 473, 496, 560, 567, 590, 654],
                        [39, 55, 73, 133, 149, 167, 227, 243, 261, 321, 337, 355,
                         415, 431, 449, 509, 525, 543, 603, 619, 637],
                        [9, 32, 83, 103, 126, 177, 197, 220, 271, 291, 314, 365,
                         385, 408, 459, 479, 502, 553, 573, 596, 647]]
        self.assertEqual(4, len(vols))
        self.assertEqual(21, len(vols[0]))
        self.assertListEqual(vols, correct_vols)

    def test_get_volumes_for_conditions_2c(self):
        conditions = [
            [("number", "d1"), ("shape", "cr")],
            [("number", "d2"), ("shape", "cr")],
            [("number", "d3"), ("shape", "cr")],
            [("number", "d5"), ("shape", "cr")]
        ]

        vols = self.s_plotter_v2.get_volumes_for_conditions(conditions)
        correct_vols = [[16, 110, 204, 298, 392, 486, 580],
                        [3, 97, 191, 285, 379, 473, 567],
                        [39, 133, 227, 321, 415, 509, 603],
                        [9, 103, 197, 291, 385, 479, 573]]
        self.assertEqual(4, len(vols))
        self.assertEqual(7, len(vols[0]))
        self.assertListEqual(vols, correct_vols)

    def test_get_volumes_for_conditions_3c(self):
        conditions = [
            [("number", "d1"), ("shape", "cr"), ("spread", "ch")],
            [("number", "d2"), ("shape", "cr"), ("spread", "ch")],
            [("number", "d3"), ("shape", "cr"), ("spread", "ch")],
            [("number", "d5"), ("shape", "cr"), ("spread", "ch")]
        ]

        vols = self.s_plotter_v2.get_volumes_for_conditions(conditions)
        correct_vols = [[16, 204, 392, 580],
                        [3, 191, 379, 567],
                        [39, 227, 415, 603],
                        [9, 197, 385, 573]]
        self.assertEqual(4, len(vols))
        self.assertEqual(4, len(vols[0]))
        self.assertListEqual(vols, correct_vols)

    def test_plot_covariates_psh_b_same(self):
        trace_id = 0
        conditions = [
            [("number", "d1")],
            [("number", "d2")],
            [("number", "d3")],
            [("number", "d5")]
        ]
        padding = [-2, -1, 0, 1, 2, 3, 4]
        _ = self.s_plotter_v2.plot_covariates_psh(None, trace_id, conditions, padding, show_plot=True)

    def test_plot_covariates_psh_b_different(self):
        trace_id = 0
        conditions = [
            [("number", "d1"), ("shape", "cr")],
            [("number", "d2"), ("shape", "cr")],
            [("number", "d3"), ("shape", "cr")],
            [("number", "d5"), ("shape", "cr")]
        ]
        padding = [-2, -1, 0, 1, 2, 3, 4]
        _ = self.s_plotter_v2.plot_covariates_psh(None, trace_id, conditions, padding, show_plot=True)

    def test_plot_covariates_psh_b_different_no_individual(self):
        trace_id = 0
        conditions = [
            [("number", "d1"), ("shape", "cr")],
            [("number", "d2"), ("shape", "cr")],
            [("number", "d3"), ("shape", "cr")],
            [("number", "d5"), ("shape", "cr")]
        ]
        padding = [-2, -1, 0, 1, 2, 3, 4]
        _ = self.s_plotter_v2.plot_covariates_psh(None, trace_id, conditions, padding,
                                                  plot_individual=False, show_plot=True)

    def test_plot_covariates_psh_0_different(self):
        trace_id = 0
        conditions = [
            [("number", "d1"), ("shape", "cr")],
            [("number", "d2"), ("shape", "cr")],
            [("number", "d3"), ("shape", "cr")],
            [("number", "d5"), ("shape", "cr")]
        ]
        padding = [0]
        _ = self.s_plotter_v2.plot_covariates_psh(None, trace_id, conditions, padding,
                                                  split=False, plot_individual=False, show_plot=True)

    def test_make_covariate_psh_figure(self):
        trace_ids = [0, 1, 2, 3, 4]
        conditions = [
            [("number", "d1"), ("shape", "cr")],
            [("number", "d2"), ("shape", "cr")],
            [("number", "d3"), ("shape", "cr")],
            [("number", "d5"), ("shape", "cr")]
        ]
        padding = [-2, -1, 0, 1, 2, 3, 4]
        _ = self.s_plotter_v2.make_covariate_psh_figure(trace_ids, conditions, padding,
                                                        "Main title",
                                                        ["0", "1", "2", "3", "4"],
                                                        show_plot=True)

    def test_crop_and_pad(self):
        trace_id = 0
        trace = self.s_plotter_v2.get_trace(trace_id)
        vols = [0, 1, 2, 3]
        padding = [-2, -1, 0, 1, 2, 3]
        cropped_trace = self.s_plotter_v2.crop_and_pad(trace, vols, padding)
        correct_crop = [[-0.00010753, -0.00010753, -0.00010753, 0.00078853, 0.00120409, 0.00697],
                        [-0.00010753, -0.00010753, 0.00078853, 0.00120409, 0.00697, 0.01724216],
                        [-0.00010753, 0.00078853, 0.00120409, 0.00697, 0.01724216, 0.00376239],
                        [0.00078853, 0.00120409, 0.00697, 0.01724216, 0.00376239, -0.00060101]]
        np.testing.assert_almost_equal(cropped_trace, correct_crop, decimal=7)


if __name__ == "__main__":
    unittest.main()



from .analysis import *
from .utils import *
from .project import *
from .notifications import *
from .plots import *
from .report import *
from .runner import *

# Version of the numan package
__version__ = "1.0.6"


from tifffile import TiffFile, imread, imsave
import numpy as np
import json
import os

import matplotlib.pyplot as plt
import warnings
from tqdm import tqdm
import pandas as pd

import PyPDF2
from PyPDF2 import PdfFileReader
from reportlab.pdfgen import canvas
from reportlab.lib.pagesizes import letter
from reportlab.lib.colors import white, black

from .analysis import *
from .utils import *


def place_cb(can, x, y, name):
    form = can.acroForm
    can.setFont("Courier", 12)
    can.drawCentredString(x + 20, y + 20, name)
    form.checkbox(name=name,
                  # tooltip = f"Field {name}",
                  x=x + 10,
                  y=y - 4,
                  # buttonStyle = 'check',
                  borderColor=black,
                  fillColor=white,
                  textColor=black,
                  forceBorder=True
                  )
    return can


def generate_timpoints(bb, ba, time_centers):
    """
    Adds the bb and ba number of blanks around each value in time_centers,
    keeping the number of rows the same.
    """
    time_points = np.zeros((3, ((bb + ba) + 1) * 3))
    for it, t in enumerate(time_centers):
        a, b, c = t
        a_long = np.concatenate((a - np.arange(bb + 1)[::-1], a + 1 + np.arange(ba)))
        b_long = np.concatenate((b - np.arange(bb + 1)[::-1], b + 1 + np.arange(ba)))
        c_long = np.concatenate((c - np.arange(bb + 1)[::-1], c + 1 + np.arange(ba)))
        time_points[it] = np.concatenate((a_long, b_long, c_long))
    time_points = time_points.astype(int)
    return time_points


def get_idx_per_page(spots, group_tag, sort_by_sig=False):
    # some info on the cells to put into the title
    cells_idx = spots.get_group_idx(spots.groups[group_tag])
    if sort_by_sig:
        cells_group = spots.get_group_info(["sig2v3", "sig2v5", "sig3v5", "sig2vB", "sig3vB", "sig5vB"],
                                           group=spots.groups[group_tag])
        cells_group = np.array([group_name.replace("sig", "") for group_name in cells_group])
        # sort everything so that the cells with the most amount of significant stuff appear first
        sorted_zip = sort_by_len0(zip(cells_group, cells_idx))
        cells_group = np.array([el[0] for el in sorted_zip])
        cells_idx = np.array([el[1] for el in sorted_zip])
    tpp = 5
    # prepare the batches per page
    cells = np.arange(len(cells_idx))
    btchs = [cells[s: s + tpp] for s in np.arange(np.ceil(len(cells_idx) / tpp).astype(int)) * tpp]
    return cells_idx.astype(str), btchs


class SignalPlotter:
    """
    All the plotting functions.
    """

    def __init__(self, signals, experiment, spf=1):
        """
        spf : seconds per frame
        """
        self.signals = signals
        self.experiment = experiment
        self.n_signals = self.signals.traces.shape[1]

    def plot_labels(self, ax, extent=None, time_points=None, front_to_tail=None):
        """
        Figures out a
        :param ax:
        :type ax:
        :param extent:
        :type extent:
        :param time_points:
        :type time_points:
        :param front_to_tail:
        :type front_to_tail:
        :return:
        :rtype:
        """
        # timing in volumes, since one volume is one time point of the signal
        timing = (self.experiment.cycle.timing / self.experiment.volume_manager.fpv).astype(int)
        # get condition name for each time point of the signal
        conditions = [cond for t, condition in zip(timing, self.experiment.cycle.conditions) for cond in
                      [condition.name] * t]

        # encode unique names into intengers, return_inverse gives the integer encoding
        names, values = np.unique(conditions, return_inverse=True)

        if time_points is not None:
            time_points = np.array(time_points)
            time_shape = time_points.shape
            assert len(time_shape) < 3, "time_shape should be 1D or 2D"
            if len(time_shape) == 2:
                time_points = time_points[0, :]
            # take only the relevant part of the condition labels
            values = values[time_points]

        if front_to_tail is not None:
            old_order = np.arange(len(values))
            new_order = np.r_[old_order[front_to_tail:], old_order[0:front_to_tail]]
            values = values[new_order]

        img = ax.imshow(values[np.newaxis, :], aspect='auto',
                        extent=extent, cmap=plt.get_cmap('Greys', len(names)))
        img.set_clim(0, len(names) - 1)

        return names, values, img

    def show_labels(self, x_step=1):
        """
        Keep in mind - assign colors in alphabetic order of the condition name.
        """
        # TODO : for now it fits 3 different colors only! fix it!
        fig = plt.figure(figsize=(10, 5))
        ax = fig.add_subplot(111)
        names, values, img = self.plot_labels(ax)
        plt.xticks(ticks=np.arange(0, len(values), x_step))
        plt.xlabel('volume # per cycle')
        plt.title('Stimulus cycle')
        ax.get_yaxis().set_visible(False)

        # TODO : for now it fits 3 different colors only! fix it!
        cbar = plt.colorbar(img, ax=ax, ticks=[0.5, 1, 1.5], orientation='horizontal')
        cbar.ax.set_xticklabels(names)

    def plot_trace(self, ax, trace,
                   error_type="sem",
                   time_points=None,
                   cycles=None,
                   front_to_tail=None,
                   plot_individual=True,
                   signal_split=None,
                   vlines=None,
                   noise_color='-c',
                   mean_color='r',
                   ax_limits=None):
        """
        Places a specified trace at the axis , ax.
        """
        # get the signals
        cycled, mean, e = self.signals.get_looped(trace, self.experiment, error_type=error_type,
                                                  time_points=time_points, cycles=cycles)
        # shift the signals front-to-tail
        if front_to_tail is not None:
            old_order = np.arange(len(mean))
            new_order = np.r_[old_order[front_to_tail:], old_order[0:front_to_tail]]
            cycled = cycled[:, new_order]
            mean = mean[new_order]
            e = e[:, new_order]

        # get axis limits
        if ax_limits is None:
            xmin, xmax, ymin, ymax = get_ax_limits(cycled, mean, e, plot_individual)
            ax_limits = (xmin, xmax, ymin, ymax)
        else:
            xmin, xmax, ymin, ymax = ax_limits

        # create the stimuli labels in the background
        names, _, img = self.plot_labels(ax, extent=[xmin, xmax, ymin, ymax],
                                         time_points=time_points,
                                         front_to_tail=front_to_tail)

        # if you wish to not connect/disconect certain groups of signals,
        # it's indexed AFTER looping and time_points were already done:
        #  index along the x axis you will see
        if signal_split is not None:
            for signal_group in signal_split:
                if plot_individual:
                    ax.plot(signal_group, cycled[:, signal_group].T, noise_color, alpha=0.3)
                plot_errorbar(ax, mean[signal_group], e[:, signal_group], x=signal_group, color=mean_color)
        else:
            if plot_individual:
                ax.plot(cycled.T, noise_color, alpha=0.4, linewidth=1)
            plot_errorbar(ax, mean, e, color=mean_color)

        # to separate the plot regions with vertical lines
        if vlines is not None:
            ax.vlines(vlines, ymin, ymax, linewidth=0.8, color='black')  # , linestyle=(0, (5, 10))

        # axis clean-up
        ax.set_xlim((xmin, xmax))
        ax.set_ylim((ymin, ymax))
        ax.set_xticks(np.arange(len(mean)))
        return ax_limits

    def show_psh(self, traces, main_title, tittles, error_type="sem",
                 time_points=None, cycles=None,
                 plot_individual=True, front_to_tail=None,
                 figure_layout=None, figsize=None,
                 ylabel='', xlabel='', noise_color='-c', vlines=None, signal_split=None,
                 gridspec_kw=None,
                 dpi=160):
        """
        Plots specified traces for all the traces.
        front_to_tail : how many cycle points to attach from front to tail
        """

        if figure_layout is not None:
            n_rows = figure_layout[0]
            n_col = figure_layout[1]
        else:
            n_rows = len(traces)
            n_col = 1

        if figsize is None:
            figsize = (12, n_rows * 4)

        fig, axes = plt.subplots(n_rows, n_col, gridspec_kw=gridspec_kw, figsize=figsize, dpi=dpi)
        axes = axes.flatten()
        fig.suptitle(main_title)

        for plot_id, trace in enumerate(traces):
            ax = axes[plot_id]
            self.plot_trace(ax, trace,
                            error_type=error_type,
                            time_points=time_points,
                            cycles=cycles,
                            front_to_tail=front_to_tail,
                            plot_individual=plot_individual,
                            signal_split=signal_split,
                            noise_color=noise_color,
                            vlines=vlines)
            ax.set_title(tittles[plot_id])
            ax.set_xticklabels([])
            ax.set_ylabel(ylabel)
            ax.set_xlabel(xlabel)

        # cbar = plt.colorbar(img, ax=ax, ticks=[0.5, 1.5, 2.5], orientation='horizontal')
        # cbar.ax.set_xticklabels(names)


class Reports:
    """
    For now it is simply a class of wrappers to make reports specifically for the 2vs3vs5 experiment.
    I hope it will become more clean and general as time goes on.
    """

    def __init__(self, project_folder, experiment):
        self.project = project_folder
        self.experiment = experiment

    @staticmethod
    def prepare_spot_info(spots, group_tag, sort_by_sig=True,
                          groups_to_specify=["sig2v3", "sig2v5", "sig3v5", "sig2vB", "sig3vB", "sig5vB"]):
        # some info on the cells to put into the title
        # cells idx in the original set of cells as they are in the spots
        cells_idx = spots.get_group_idx(spots.groups[group_tag])
        # cells coordinates
        cells_zyx = spots.get_group_centers(spots.groups[group_tag]).astype(np.int32)
        # list of groups that the cell belongs to ( from the list provided )
        cells_group = spots.get_group_info(groups_to_specify,
                                           group=spots.groups[group_tag])
        cells_group = np.array([group_name.replace("sig", "") for group_name in cells_group])
        # since signals are already cropped for the group, it's just np.arrange
        signal_idx = np.arange(sp.n_signals)

        if sort_by_sig:
            # sort everything so that the cells with the most amount of significant stuff appear first
            sorted_zip = sort_by_len0(zip(cells_group, cells_idx, cells_zyx, signal_idx))
            cells_group = np.array([el[0] for el in sorted_zip])
            cells_idx = np.array([el[1] for el in sorted_zip])
            cells_zyx = np.array([el[2] for el in sorted_zip])
            signal_idx = np.array([el[3] for el in sorted_zip])

        return cells_group, cells_idx, cells_zyx, signal_idx

    def make_signal_reports(self, spot_tag, group_tag,
                            plot_type="cycle",
                            plot_type_tag='',
                            front_to_tail=0,
                            time_points=None,
                            vlines=None,
                            signal_split=None,
                            error_type="sem",
                            noise_color='--c',
                            plot_individual=False,
                            tmp_folder=None,
                            pdf_filename=None):
        """
        Generates a pdf with the specified type of plots.

        :param spot_tag: what set of spots to use. Chooses the spots*.json based on this.
        :type spot_tag: str
        :param group_tag: what spots group to use.
        :type group_tag: str
        :param plot_type: what plot to output ["cycle","psh_0","psh_b"]
        :type plot_type: str
        :param plot_type_tag: just for the pdf naming : this is to be able to distinguish
                    the pdfs with the same plot type, but errors are different or raw traces on/off or front_to_tail...
        :type plot_type_tag: str
        :param front_to_tail: front_to_tail will shift the cycle by the set number of voxels
                    so when set to 3, there are 3 blank volumes at the begining and at the end ...
                    if set to 0, will have 6 leading blanks and will end right after the 5 dots (black bar)
        :type front_to_tail: int
        :param time_points: only show certain timepoints from the signal, for example : only 2 dots.
                    IF time_points is 2d array, will overlap traces along axis = 1.
        :type time_points: numpy.array
        :param vlines: draw vertical lines, locations (in volumes) where to draw vertical lines
        :type vlines: list[int]
        :param signal_split: how to break the lines in the plot, this is relevant to the displayed x axis
        :type signal_split: numpy.array
        :param error_type: what error type to use ( "sem" for SEM or "prc" for 5th - 95th percentile )
        :type error_type: str
        :param noise_color: the color of the individual traces (if shown)
        :type noise_color: valid color definition
        :param plot_individual: wheather to plot the individual traces
        :type plot_individual: bool
        :param tmp_folder: will store batch images in this folder before merging pdf,
                    will be stored with reports if left None.
        :type tmp_folder: Union(str, Path)
        :param pdf_filename: the name of the pdf file to save, will be generated automatically if left None
        :type pdf_filename: str

        """

        # fill out the defaults
        if tmp_folder is None:
            # where to temporary store images while the code is running
            tmp_folder = f"{self.project}/spots/reports/all_significant/signals/"
        if pdf_filename is None:
            # filename to save pdf with all the significant traces
            pdf_filename = f"{self.project}/spots/reports/all_significant/signals/" \
                           f"{plot_type}{plot_type_tag}_from_{spot_tag}_significance_{group_tag}.pdf"

        spots = Spots.from_json(f"{self.project}/spots/signals/spots_{spot_tag}.json")

        # initialise the signal plotter with DFF signal
        SLIDING_WINDOW = 15  # in volumes
        print(f"Using sliding window {SLIDING_WINDOW} volumes for signal DFF")
        significant_signals_dff = spots.get_group_signals(spots.groups[group_tag]).as_dff(SLIDING_WINDOW)
        sp = SignalPlotter(significant_signals_dff, self.experiment)

        # prepare title info
        cells_group, cells_idx, cells_zyx, signal_idx = self.prepare_spot_info(spots, group_tag, sort_by_sig=True,
                          groups_to_specify=["sig2v3", "sig2v5", "sig3v5", "sig2vB", "sig3vB", "sig5vB"])
        main_title = f"DFF signals, tscore image {spot_tag}, significance {group_tag}"

        # choose traces per page
        if plot_type == "psh_0":
            tpp = 10  # traces per page
        else:
            tpp = 5

        # prepare the batches per page
        cells = np.arange(sp.n_signals)
        btchs = [cells[s: s + tpp] for s in np.arange(np.ceil(sp.n_signals / tpp).astype(int)) * tpp]

        pdfs = []

        for ibtch, btch in enumerate(btchs):

            if plot_type == "cycle":
                # titles for the current batch
                titles = [f"Cell {idx}, {group} XYZ : {zyx[2]},{zyx[1]},{zyx[0]} (voxel) "
                          for idx, group, zyx in zip(cells_idx[btch], cells_group[btch], cells_zyx[btch])]

                sp.show_psh(signal_idx[btch],
                            main_title,
                            titles,
                            # front_to_tail will shift the cycleby the set number of voxels
                            # so when set to 3, there are 3 blank volumes at the begining and at the end ...
                            # if set to 0, will have 6 leading blanks and will end right after the 5 dots (black bar)
                            front_to_tail=front_to_tail,
                            # what grid to use to show the points
                            figure_layout=[5, 1],
                            # what error type to use ( "sem" for SEM or "prc" for 5th - 95th percentile )
                            error_type=error_type,
                            # figure parameters
                            figsize=(10, 12),
                            dpi=60,
                            # wheather to plot the individual traces
                            plot_individual=plot_individual,
                            # the color of the individual traces (if shown)
                            noise_color=noise_color)

            if plot_type == "psh_0":
                # titles for the current batch
                titles = [f"Cell {idx}, {group} \nXYZ : {zyx[2]},{zyx[1]},{zyx[0]} (voxel) "
                          for idx, group, zyx in zip(cells_idx[btch], cells_group[btch], cells_zyx[btch])]
                sp.show_psh(signal_idx[btch],
                            main_title,
                            titles,
                            # only show certain timepoints from the signal, for example : only 2 dots
                            time_points=time_points,
                            # front_to_tail will shift the cycleby the set number of voxels
                            # so when set to 3, there are 3 blank volumes at the begining and at the end ...
                            # if set to 0, will have 6 leading blanks and will end right after the 5 dots (black bar)
                            front_to_tail=0,
                            # what grid to use to show the points
                            figure_layout=[5, 2],
                            # what error type to use ( "sem" for SEM or "prc" for 5th - 95th percentile )
                            error_type=error_type,
                            # figure parameters
                            figsize=(10, 12),
                            dpi=60,
                            gridspec_kw={'hspace': 0.4, 'wspace': 0.3},
                            # wheather to plot the individual traces
                            plot_individual=plot_individual,
                            # the color of the individual traces (if shown)
                            noise_color=noise_color)

            if plot_type == "psh_b":
                # titles for the current batch
                titles = [f"Cell {idx}, {group} XYZ : {zyx[2]},{zyx[1]},{zyx[0]} (voxel) "
                          for idx, group, zyx in zip(cells_idx[btch], cells_group[btch], cells_zyx[btch])]
                sp.show_psh(signal_idx[btch],
                            main_title,
                            titles,
                            # only show certain timepoints from the signal, for example : only 2 dots
                            time_points=time_points,
                            # front_to_tail will shift the cycleby the set number of voxels
                            # so when set to 3, there are 3 blank volumes at the begining and at the end ...
                            # if set to 0, will have 6 leading blanks and will end right after the 5 dots (black bar)
                            front_to_tail=0,
                            # what grid to use to show the points
                            figure_layout=[5, 1],
                            # what error type to use ( "sem" for SEM or "prc" for 5th - 95th percentile )
                            error_type="sem",
                            # figure parameters
                            figsize=(10, 12),
                            dpi=60,
                            # if you wish to split the line
                            signal_split=signal_split,
                            # wheather to plot the individual traces
                            plot_individual=plot_individual,
                            # if you want to add vertical lines anywhere, list the x locations
                            vlines=vlines,
                            # the color of the individual traces (if shown)
                            noise_color=noise_color)

            plt.xlabel('Volume in cycle')
            filename = f'{tmp_folder}signals_batch{ibtch}.pdf'
            plt.savefig(filename)
            plt.close()
            pdfs.append(filename)

        merge_pdfs(pdfs, pdf_filename)

    def make_group_selection(self, bb=3, ba=5, time_centers=None, spot_tag=None,
                             plot_type="psh_b", plot_individual=False, rewrite=False):

        """
        Creates files with the checkboxes for group selection. The group tag is set to "sigAny2v3v5vB" fo rnow...
        """
        group_tag = "sigAny2v3v5vB"
        sort_by_sig = True
        spots = Spots.from_json(f"{self.project}/spots/signals/spots_{spot_tag}.json")

        tmp_folder = f"{self.project}/spots/reports/groupped/tmp/{spot_tag}_{group_tag}_png"

        if os.path.isdir(tmp_folder) and not rewrite:
            print(
                "The folder with the images already exists, reusing existing images. Do you want to rewrite the images? ( set rewrite = True)")
        else:
            os.makedirs(tmp_folder)

            if plot_type == "psh_b":
                # time points for the psh_b
                time_points = generate_timpoints(bb, ba, time_centers)
                time_points[2, -ba:] = np.arange(ba)
                # create a way to break the lines on the plot (to visually separate different stimuli)
                signal_split = np.array([np.arange(bb + ba + 1),
                                         np.arange(bb + ba + 1) + (
                                                 bb + ba + 1),
                                         np.arange(bb + ba + 1) + (
                                                 bb + ba + 1) * 2])
                vlines = [8.5, 17.5]
            else:
                time_points = None
                signal_split = None
                vlines = None

            self.prepare_group_images(spots, spot_tag, group_tag, tmp_folder,
                                      # types of plots:
                                      plot_type=plot_type,
                                      dpi=160,
                                      # just for the pdf naming :
                                      # this is to be able to distinguish the pdfs with the same plot type,
                                      # but errors are different or raw traces on/off or front_to_tail
                                      plot_type_tag='',
                                      # only show certain timepoints from the signal, for example : only 2 dots
                                      time_points=time_points,
                                      # how to break the line
                                      signal_split=signal_split,
                                      # draw vertical lines
                                      vlines=vlines,
                                      # wether or not you want to have the cells sorted on how many tests they passes
                                      sort_by_sig=sort_by_sig,
                                      # what error type to use ( "sem" for SEM or "prc" for 5th - 95th percentile )
                                      error_type="sem",
                                      # wheather to plot the individual traces
                                      plot_individual=plot_individual,
                                      # the color of the individual traces (if shown)
                                      noise_color='-c')

        # get cell idx per page
        cells_idx, btchs = get_idx_per_page(spots, group_tag, sort_by_sig=sort_by_sig)

        # create pdf
        can = canvas.Canvas(f"spots/reports/groupped/from_{spot_tag}_significance_{group_tag}_choose.pdf",
                            pagesize=letter)
        # for each page
        for ibtch, btch in enumerate(btchs):
            # refresh checkbox locations to be at the top, numbers fitted to match the plots in the shape (5,1)
            X, Y, H = 10, 600, 122
            # add image
            can.drawImage(f"{tmp_folder}/signals_batch{ibtch}.png", 0, -600, width=650,
                          preserveAspectRatio=True, mask='auto')
            # add checkboxes
            for cell_name in cells_idx[btch]:
                can = place_cb(can, x=X, y=Y, name=cell_name)
                Y = Y - H
            # finish page
            can.showPage()
        can.save()

    def prepare_group_images(self, spots, spot_tag, group_tag, tmp_folder,
                             plot_type="cycle",
                             plot_type_tag='',
                             dpi=160,
                             front_to_tail=0,
                             time_points=None,
                             vlines=None,
                             signal_split=None,
                             error_type="sem",
                             noise_color='--c',
                             plot_individual=False):
        """
        plot_type: "psh_b" or "cycle" only
        """
        # initialise the signal plotter
        SLIDING_WINDOW = 15  # in volumes
        significant_signals_dff = spots.get_group_signals(spots.groups[group_tag]).as_dff(SLIDING_WINDOW)
        sp = SignalPlotter(significant_signals_dff, self.experiment)

        # some info on the cells to put into the title
        cells_group, cells_idx, cells_zyx, signal_idx = self.prepare_spot_info(spots, group_tag, sort_by_sig=True,
                                                                               groups_to_specify=["sig2v3", "sig2v5",
                                                                                                  "sig3v5", "sig2vB",
                                                                                                  "sig3vB", "sig5vB"])
        main_title = f"DFF signals, tscore image {spot_tag}, significance {group_tag}"

        if plot_type == "psh_0":
            tpp = 10  # traces per page
        else:
            tpp = 5
        # prepare the batches per page
        cells = np.arange(sp.n_signals)
        btchs = [cells[s: s + tpp] for s in np.arange(np.ceil(sp.n_signals / tpp).astype(int)) * tpp]

        for ibtch, btch in enumerate(btchs):

            if plot_type == "cycle":
                # titles for the current batch
                titles = [f"Cell {idx}, {group} XYZ : {zyx[2]},{zyx[1]},{zyx[0]} (voxel) "
                          for idx, group, zyx in zip(cells_idx[btch], cells_group[btch], cells_zyx[btch])]

                sp.show_psh(signal_idx[btch],
                            main_title,
                            titles,
                            # front_to_tail will shift the cycleby the set number of voxels
                            # so when set to 3, there are 3 blank volumes at the begining and at the end ...
                            # if set to 0, will have 6 leading blanks and will end right after the 5 dots (black bar)
                            front_to_tail=front_to_tail,
                            # what grid to use to show the points
                            figure_layout=[5, 1],
                            # what error type to use ( "sem" for SEM or "prc" for 5th - 95th percentile )
                            error_type=error_type,
                            # figure parameters
                            figsize=(10, 12),
                            dpi=dpi,
                            # wheather to plot the individual traces
                            plot_individual=plot_individual,
                            # the color of the individual traces (if shown)
                            noise_color=noise_color)

            if plot_type == "psh_b":
                # titles for the current batch
                titles = [f"Cell {idx}, {group} XYZ : {zyx[2]},{zyx[1]},{zyx[0]} (voxel) "
                          for idx, group, zyx in zip(cells_idx[btch], cells_group[btch], cells_zyx[btch])]
                sp.show_psh(signal_idx[btch],
                            main_title,
                            titles,
                            # only show certain timepoints from the signal, for example : only 2 dots
                            time_points=time_points,
                            # front_to_tail will shift the cycleby the set number of voxels
                            # so when set to 3, there are 3 blank volumes at the begining and at the end ...
                            # if set to 0, will have 6 leading blanks and will end right after the 5 dots (black bar)
                            front_to_tail=0,
                            # what grid to use to show the points
                            figure_layout=[5, 1],
                            # what error type to use ( "sem" for SEM or "prc" for 5th - 95th percentile )
                            error_type="sem",
                            # figure parameters
                            figsize=(10, 12),
                            dpi=dpi,
                            # if you wish to split the line
                            signal_split=signal_split,
                            # wheather to plot the individual traces
                            plot_individual=plot_individual,
                            # if you want to add vertical lines anywhere, list the x locations
                            vlines=vlines,
                            # the color of the individual traces (if shown)
                            noise_color=noise_color)

            plt.xlabel('Volume in cycle')
            filename = f'{tmp_folder}/signals_batch{ibtch}.png'
            plt.savefig(filename)
            plt.close()



from tifffile import TiffFile, imread, imsave
import numpy as np
import json
import os

import matplotlib.pyplot as plt
import warnings
from tqdm import tqdm
import pandas as pd

from itertools import combinations

import PyPDF2
from PyPDF2 import PdfFileReader
from reportlab.pdfgen import canvas
from reportlab.lib.pagesizes import letter
from reportlab.lib.colors import white, black
from scipy.stats import rankdata

# from .analysis import *
# from .utils import *
from numan import *


# def place_cb(can, x, y, name):
#     form = can.acroForm
#     can.setFont("Courier", 12)
#     can.drawCentredString(x + 20, y + 20, name)
#     form.checkbox(name=name,
#                   # tooltip = f"Field {name}",
#                   x=x + 10,
#                   y=y - 4,
#                   # buttonStyle = 'check',
#                   borderColor=black,
#                   fillColor=white,
#                   textColor=black,
#                   forceBorder=True
#                   )
#     return can
#

# def generate_timpoints(bb, ba, time_centers):
#     """
#     Adds the bb and ba number of blanks around each value in time_centers,
#     keeping the number of rows the same.
#     """
#     time_points = np.zeros((3, ((bb + ba) + 1) * 3))
#     for it, t in enumerate(time_centers):
#         a, b, c = t
#         a_long = np.concatenate((a - np.arange(bb + 1)[::-1], a + 1 + np.arange(ba)))
#         b_long = np.concatenate((b - np.arange(bb + 1)[::-1], b + 1 + np.arange(ba)))
#         c_long = np.concatenate((c - np.arange(bb + 1)[::-1], c + 1 + np.arange(ba)))
#         time_points[it] = np.concatenate((a_long, b_long, c_long))
#     time_points = time_points.astype(int)
#     return time_points


def shift_signal(values, forward_shift, axis=0):
    """
    Shifts signal along axis 0 by forward_shift number of units forward. ( Tail will be in the front )
    :param axis: 0 or -1 , whether to reorder the o dimension or the last one
    :type axis: int, 0/-1
    :param values: values to shift
    :type values: Union(list, numpy.array)
    :param forward_shift: how many units to shift forward by
    :type forward_shift: int
    :return: shifted signal
    :rtype: Union(list, numpy.array)
    """
    assert axis in [0, -1], f"axis should be 0 or -1, but {axis} was given"

    old_order = np.arange(values.shape[axis])
    new_order = np.r_[old_order[forward_shift:], old_order[0:forward_shift]]
    if axis == 0:
        values = values[new_order]
    elif axis == -1:
        values = values[:, new_order]
    return values


def select_columns(values, selection):
    """
    Selects only certain columns from the given array.
    :param values: array to select values from, can be 1d or 2d
    :type values: Union(list, numpy.array)
    :param selection: positions along axis 1 of the elements to grab
    :type selection: [int]
    :return: cropped
    :rtype: Union(list, numpy.array)
    """
    selection = np.array(selection)
    # returns 1d ("cropped") or 2d ("looped")
    selection_dim = len(selection.shape)
    values_dim = len(values.shape)

    if values_dim == 1:
        values = values[selection]

    # returns 2d ("cropped" or "looped")
    elif values_dim == 2:
        values = values[:, selection]
        if selection_dim == 2:
            values = values.reshape(-1, values.shape[2])

    return values


# def get_idx_per_page(spots, group_tag, sort_by_sig=False):
#     # some info on the cells to put into the title
#     cells_idx = spots.get_group_idx(spots.groups[group_tag])
#     if sort_by_sig:
#         cells_group = spots.get_group_info(["sig2v3", "sig2v5", "sig3v5", "sig2vB", "sig3vB", "sig5vB"],
#                                            group=spots.groups[group_tag])
#         cells_group = np.array([group_name.replace("sig", "") for group_name in cells_group])
#         # sort everything so that the cells with the most amount of significant stuff appear first
#         sorted_zip = sort_by_len0(zip(cells_group, cells_idx))
#         cells_group = np.array([el[0] for el in sorted_zip])
#         cells_idx = np.array([el[1] for el in sorted_zip])
#     tpp = 5
#     # prepare the batches per page
#     cells = np.arange(len(cells_idx))
#     btchs = [cells[s: s + tpp] for s in np.arange(np.ceil(len(cells_idx) / tpp).astype(int)) * tpp]
#     return cells_idx.astype(str), btchs


class LabelPlotter:
    """
    Plots conditions for one cycle or a portion of a cycle.
    """

    # TODO : add per frame possibility

    def __init__(self, experiment, annotation_type):
        """Only per volume labels are implemented at the moment"""
        self.experiment = experiment
        # get conditions per volumes
        self.names, self.values = self.conditions_per_cycle(annotation_type)

    def conditions_per_cycle(self, annotation_type):
        """
        Returns names of labels in one cycle and labels per volume for one cycle.
        :param annotation_type: the name of the annotation to list
        :type annotation_type: str
        :return: a list of label names that occur in one cycle
        and a list of label's ranks per volume in a cycle : [0,0,0,3, 0, 0 ...0, 1, 0, ..]
        :rtype: ([str],[int])
        """
        condition_ids, condition_names = self.experiment.list_conditions_per_cycle(annotation_type)
        # make it so that the conditions start at 0 and grow incrementally : 0, 1, 2, 3, ...
        conditions = np.unique(condition_ids)
        condition_rank = [np.where(conditions == c_id)[0][0] for c_id in condition_ids]
        # keep only the names of the conditions present:
        # some conditions might have been listed as possible conditions, but not used in annotation
        # (also numbering cond starts at 1, so "cond-1" )
        names = [condition_names[cond - 1] for cond in conditions]

        return names, condition_rank

    def plot_labels(self, ax=None, extent=None, time_points=None, forward_shift=None, show_plot=False):
        """
        Figures out how to plot labels.
        :param ax:
        :type ax:
        :param extent:
        :type extent:
        :param time_points: the columns from the data to keep
        :type time_points: Union([int], numpy.array(int))
        :param forward_shift:
        :type forward_shift:
        :param show_plot: whether to show the plotted axis
        :type show_plot: bool
        :return:
        :rtype:
        """
        if ax is None:
            ax = plt.gca()

        values = np.array(self.values)

        if time_points is not None:
            assert len(np.array(time_points).shape) == 1, f"time_points for labels have to be a 1d array," \
                                                          f"but got {time_points.shape}"
            # take only the relevant part of the condition labels
            values = select_columns(values, time_points)

        if forward_shift is not None:
            values = shift_signal(values, forward_shift, axis=0)

        img = ax.imshow(values[np.newaxis, :], aspect='auto',
                        extent=extent, cmap=plt.get_cmap('Greys', len(self.names)))
        img.set_clim(0, len(self.names) - 1)

        if show_plot:
            plt.show()

        return self.names, values, img


class SignalPlotter:
    """
    Plots signals per VOLUMES.
    """

    def __init__(self, signals, experiment, annotation_type,
                 mean_color='r', noise_color='-c',
                 # covariate mean and noise
                 c_mean_color='w', c_noise_color='-m', c_edge_color='w'):

        self.traces = signals.traces  # Time x N_traces
        self.n_traces = self.traces.shape[1]

        self.experiment = experiment
        self.annotation = annotation_type
        self.labels = LabelPlotter(experiment, annotation_type)

        # plotting parameters
        self.mean_color = mean_color
        self.noise_color = noise_color
        self.c_mean_color = c_mean_color
        self.c_noise_color = c_noise_color
        self.c_edge_color = c_edge_color

        self.error_type = "sem"

    def get_trace(self, trace_id):
        return self.traces[:, trace_id]

    def prepare_cycle(self, trace):
        """
        Preppares trace to be plotted as cycle.
        """
        cycle_iterations = self.experiment.list_cycle_iterations(self.annotation, as_volumes=True)
        cycles, volumes_per_cycle = np.unique(cycle_iterations, return_counts=True)
        trace = trace.reshape((len(cycles), volumes_per_cycle[0]))
        return trace

    @staticmethod
    def pad_cycle(trace, padding):
        pad_left = min(padding)
        # move everything down a cycle , so the end of cycle one appears at the beginning of cycle 2
        # NOTE: the end of the VERY last cycle appears at the very beginning of the recording
        trace_left = trace[0:-1, pad_left:]
        trace_left = np.vstack((trace[-1, pad_left:][None, :], trace_left))

        pad_right = max(padding)
        # move everything up a cycle , so the beginning of cycle 2 appears at the end of cycle 1
        # NOTE: the beginning of the VERY first cycle appears at the end of the recording
        trace_right = trace[1:, 0:pad_right]
        trace_right = np.vstack((trace_right, trace[0, 0:pad_right][None, :]))

        return np.hstack((trace_left, trace, trace_right))

    @staticmethod
    def crop_and_pad(trace, volumes, padding):
        """

        :param trace: signal trace from which to crop the time points
        :type trace: 1d numpy array
        :param volumes: list of central volumes to crop from the signal, of length V
        :type volumes: [int]
        :param padding: padding around the central volumes, of length P
        :type padding: [int]
        :return: cropped signal in a shape VxP
        :rtype: 2d numpy array
        """
        padding = np.array(padding)[np.newaxis, :]
        volumes = np.array(volumes)[:, np.newaxis]

        assert np.min(padding) <= 0, \
            f"Padding on the left should be negative or 0, but got {min(padding)}"
        assert np.max(padding) >= 0, \
            f"Padding on the right should be positive or 0, but got {max(padding)}"

        # will make a matrix VxP
        mask = volumes + padding
        # make sure you are not requesting things outside the trace
        # this will copy the values at the beginning/ end of the trace
        mask[mask < 0] = 0
        mask[mask >= len(trace)] = len(trace) - 1

        cropped_signal = trace[mask]
        return cropped_signal

    @staticmethod
    def get_trace_stats(trace, error_type="sem"):
        """
        Calculates trace statistics along axis = 0.
        Trace should be reshaped prior to calling this function.
        """
        mean = np.mean(trace, axis=0)
        if error_type == "prc":
            # error bars : 5 to 95 th percentile around the median
            e = np.r_[np.expand_dims(mean - np.percentile(trace, 5, axis=0), axis=0),
                      np.expand_dims(np.percentile(trace, 95, axis=0) - mean, axis=0)]
        elif error_type == "sem":
            # error bars : sem around hte mean
            sem = np.std(trace, axis=0, ddof=1) / np.sqrt(trace.shape[0])
            e = np.r_[np.expand_dims(sem, axis=0),
                      np.expand_dims(sem, axis=0)]
        else:
            e = None

        return mean, e

    def get_labels_in_cycle(self, labels, padding=[0], shift_by_padding=False):
        """
        :param shift_by_padding: Whether to shift the labels by the size of padding on the left.
        Used when the signal has had a padding added on the two sides.
        :type shift_by_padding: bool
        :param labels: list of the label names to grab, from the annotation
        :type labels: [str]
        :param padding: list of padding around each label's spot : negative to the left, positive to the right
        Example: [-2,-1,0,1,2,3]
        :type padding: [int]
        :return: list of the labels locations with the padding. Concatenated in the provided order.
        :rtype: numpy.array
        """
        # figure out what columns you want from the padded cycle
        label_in_cycle = {}
        for label in labels:
            # get label location in cycle
            label_id = np.where(np.array(self.labels.names) == label)[0]
            label_in_cycle[label] = np.where(self.labels.values == label_id)[0]
            # turn into Nx1 and 1xM for broadcasting to work
            label_in_cycle[label] = label_in_cycle[label][:, None] + np.array(padding)[None, :]
            if shift_by_padding:
                # shift by the amount of padding on the left
                label_in_cycle[label] = label_in_cycle[label] + abs(min(padding))
        selection = np.hstack([label_in_cycle[label] for label in labels])
        return selection

    def prepare_psh(self, trace, labels, padding):
        """
        Prepares signal for plotting.
        """
        # TODO : check padding format
        # prepare cycles
        trace = self.prepare_cycle(trace)
        n_cycles = trace.shape[0]
        cycle_length = trace.shape[1]

        # get the labels
        for label in labels:
            assert label in self.labels.names, f"There is no label {label} " \
                                               f"in annotation {self.annotation}"

        # pad the cycle
        trace = self.pad_cycle(trace, padding)

        selection = self.get_labels_in_cycle(labels, padding=padding, shift_by_padding=True)
        trace = select_columns(trace, selection)

        # subtract min padding to get to the unpadded values on the left
        # mod to get the unpadded (and looped) values on the right
        label_selection = np.mod(selection[0] - abs(min(padding)), cycle_length)

        return trace, label_selection

    @staticmethod
    def get_signal_split(labels, padding):
        n_padding = len(padding)

        split = []
        for i_label, _ in enumerate(labels):
            split.append(np.arange(n_padding) + n_padding * i_label)

        return split

    @staticmethod
    def get_vlines(labels, padding):
        n_padding = len(padding)

        vlines = []
        for i_label, _ in enumerate(labels):
            vlines.append(n_padding - 0.5 + n_padding * i_label)

        return vlines

    def plot_trace(self, ax, trace, mean, e, plot_individual, signal_split,
                   covariate=False, show_plot=False):
        """


        :param covariate: wether or not it's plotting covariate trace
        :type covariate: bool
        :param ax: Axis to which to add the trace
        :type ax:
        :param trace: Trace to plot, must be already reshaped to the desired shape
        :type trace:
        :param mean: Mean trace
        :type mean:
        :param e: Error bars to add
        :type e:
        :param plot_individual: Whether to plot the traces of the individual runs
        :type plot_individual:
        :param signal_split: How to split the trace along the x-axis
        :type signal_split:
        :param show_plot: Whether to show the plot
        :type show_plot:
        """
        # if you wish to not connect/disconect certain groups of signals,
        # it's indexed AFTER looping and time_points were already done:
        #  index along the x axis you will see
        if ax is None:
            ax = plt.gca()

        if covariate:
            noise_color = self.c_noise_color
            mean_color = self.c_mean_color
            edge_color = self.c_edge_color
        else:
            noise_color = self.noise_color
            mean_color = self.mean_color
            edge_color = None

        if signal_split is None:
            if plot_individual:
                ax.plot(trace.T, noise_color, alpha=0.4, linewidth=1)
            plot_errorbar(ax, mean, e, color=mean_color, edge_color=edge_color)
        else:
            for signal_group in signal_split:
                if plot_individual:
                    ax.plot(signal_group, trace[:, signal_group].T, noise_color, alpha=0.3)
                plot_errorbar(ax, mean[signal_group], e[:, signal_group],
                              x=signal_group, color=mean_color, edge_color=edge_color)

        if show_plot:
            plt.show()

    def plot_cycles(self, ax, trace_id,
                    forward_shift=None,
                    plot_individual=True,
                    ax_limits=None,
                    show_plot=False):
        """
        Places a specified trace (the signal from one cell) at the provided axis , ax.
        With the corresponding labels.
        """
        if ax is None:
            ax = plt.gca()

        # get the individual signal trace
        trace = self.get_trace(trace_id)

        # get the signals to plot in the desired shape
        trace = self.prepare_cycle(trace)
        mean, e = self.get_trace_stats(trace, error_type=self.error_type)

        # shift the signals forward
        if forward_shift is not None:
            trace = shift_signal(trace, forward_shift, axis=-1)
            mean = shift_signal(mean, forward_shift, axis=0)
            e = shift_signal(e, forward_shift, axis=-1)

        # get axis limits
        if ax_limits is None:
            xmin, xmax, ymin, ymax = get_ax_limits(trace, mean, e, plot_individual)
            ax_limits = [xmin, xmax, ymin, ymax]

        # plot the trace
        self.plot_trace(ax, trace, mean, e, plot_individual, signal_split=None)

        # create the stimuli labels in the background
        names, _, img = self.labels.plot_labels(ax,
                                                forward_shift=forward_shift,
                                                extent=ax_limits)

        # axis clean-up
        xmin, xmax, ymin, ymax = ax_limits
        ax.set_xlim((xmin, xmax))
        ax.set_ylim((ymin, ymax))
        ax.set_xticks(np.arange(len(mean)))

        if show_plot:
            plt.show()

        return ax_limits

    def get_labels_for_conditions(self, conditions):
        """
        Extracts the label of the type annotation from the provided list of conditions
        :param conditions: conditions to plot, list of lists of tuples ... like this:
        [
        [(label_type,label), (label_type,label), ...],
        [(label_type,label), (label_type,label),...],
        ...
        ]
        Note: if it's a tuple of length 1, you need to write it as ("x",) not ("x") !
        :type conditions: [tup]
        :return: list of labels for each condition
        :rtype: []
        """
        condition_label = [''] * len(conditions)
        for ic, condition in enumerate(conditions):
            for label in condition:
                if label[1] in self.labels.names:
                    condition_label[ic] = label[1]
            if condition_label[ic] == "":
                raise ValueError(
                    f"condition {condition} doesn't contain a label of type {self.annotation}.")
        return condition_label

    def get_volumes_for_conditions(self, conditions):
        """
        Returns the volumes that correspond to the conditoins
        :return: list of volumes for each condition list
        :rtype:[[int]]
        """
        volumes = []
        for condition in conditions:
            volumes.append(self.experiment.choose_volumes(condition))
        return volumes

    def plot_psh(self, ax, trace_id,
                 labels,
                 padding,
                 plot_individual=True,
                 split=True,
                 ax_limits=None,
                 show_plot=False):
        """
        Places a specified trace (the signal from one cell) at the provided axis , ax.
        With the corresponding labels.
        """
        if ax is None:
            ax = plt.gca()

        # get the individual signal trace
        trace = self.get_trace(trace_id)

        # get the signals to plot in the desired shape
        trace, label_selection = self.prepare_psh(trace, labels, padding)
        mean, e = self.get_trace_stats(trace, error_type=self.error_type)

        # get axis limits
        if ax_limits is None:
            xmin, xmax, ymin, ymax = get_ax_limits(trace, mean, e, plot_individual)
            ax_limits = (xmin, xmax, ymin, ymax)
        else:
            xmin, xmax, ymin, ymax = ax_limits

        # plot the trace
        if split:
            signal_split = self.get_signal_split(labels, padding)
        else:
            signal_split = None
        self.plot_trace(ax, trace, mean, e, plot_individual, signal_split)

        # create the stimuli labels in the background
        names, _, img = self.labels.plot_labels(ax,
                                                extent=[xmin, xmax, ymin, ymax],
                                                time_points=label_selection)

        # to separate the plot regions with vertical lines
        vlines = self.get_vlines(labels, padding)
        ax.vlines(vlines, ymin, ymax, linewidth=0.8, color='black')  # , linestyle=(0, (5, 10))

        # axis clean-up
        ax.set_xlim((xmin, xmax))
        ax.set_ylim((ymin, ymax))
        ax.set_xticks(np.arange(len(mean)))

        if show_plot:
            plt.show()

        return ax_limits

    def prepare_condition_trace(self, trace, conditions, padding):
        """
        Crops the signal that corresponds to the conditions
        :return: signal for each condition, concatenated
        :rtype: numpy array
        """
        labels = self.get_labels_for_conditions(conditions)
        # get the volumes to use for each condition list
        condition_vols = self.get_volumes_for_conditions(conditions)
        # for each group of volumes, prepare the signals:
        condition_signals = []
        for label, vols in zip(labels, condition_vols):
            condition_signals.append(self.crop_and_pad(trace, vols, padding))
        return np.concatenate(condition_signals, axis=1)

    def plot_covariates_psh(self, ax, trace_id,
                            conditions,
                            padding,
                            plot_individual=True,
                            split=True,
                            ax_limits=None,
                            show_plot=False):
        """
        Places a specified trace (the signal from one cell) at the provided axis , ax.
        With the corresponding labels.
        """
        if ax is None:
            ax = plt.gca()

        # get the individual signal trace
        trace = self.get_trace(trace_id)

        # get the labels according to the annotation to use
        labels = self.get_labels_for_conditions(conditions)

        # prepare stuff for the condition trace
        condition_signals = self.prepare_condition_trace(trace, conditions, padding)
        condition_mean, condition_e = self.get_trace_stats(condition_signals, error_type=self.error_type)

        # prepare stuff for the original trace
        trace, label_selection = self.prepare_psh(trace, labels, padding)
        mean, e = self.get_trace_stats(trace, error_type=self.error_type)

        # get axis limits
        if ax_limits is None:
            xmin1, xmax1, ymin1, ymax1 = get_ax_limits(trace, mean, e, plot_individual)
            xmin2, xmax2, ymin2, ymax2 = get_ax_limits(condition_signals, condition_mean, condition_e, plot_individual)
            xmin, xmax, ymin, ymax = min(xmin1, xmin2), max(xmax1, xmax2), min(ymin1, ymin2), max(ymax1, ymax2)
            ax_limits = (xmin, xmax, ymin, ymax)
        else:
            xmin, xmax, ymin, ymax = ax_limits

        # plot the trace
        if split:
            signal_split = self.get_signal_split(labels, padding)
        else:
            signal_split = None

        self.plot_trace(ax, trace, mean, e, plot_individual, signal_split)
        self.plot_trace(ax, condition_signals, condition_mean, condition_e, plot_individual, signal_split,
                        covariate=True)

        # create the stimuli labels in the background
        names, _, img = self.labels.plot_labels(ax,
                                                extent=[xmin, xmax, ymin, ymax],
                                                time_points=label_selection)

        # to separate the plot regions with vertical lines
        vlines = self.get_vlines(labels, padding)
        ax.vlines(vlines, ymin, ymax, linewidth=0.8, color='black')  # , linestyle=(0, (5, 10))

        # axis clean-up
        ax.set_xlim((xmin, xmax))
        ax.set_ylim((ymin, ymax))
        ax.set_xticks(np.arange(len(mean)))

        if show_plot:
            plt.show()

        return ax_limits

    def make_psh_figure(self, traces, labels, padding,
                        main_title,
                        tittles,
                        plot_individual=False,
                        split=True,
                        figure_layout=None,
                        figsize=None,
                        gridspec_kw=None,
                        dpi=160,
                        show_plot=False):
        """
        Plots specified traces for all the traces.
        front_to_tail : how many cycle points to attach from front to tail
        """

        if figure_layout is not None:
            n_rows = figure_layout[0]
            n_col = figure_layout[1]
        else:
            n_rows = len(traces)
            n_col = 1
        if figsize is None:
            figsize = (12, n_rows * 4)

        fig, axes = plt.subplots(n_rows, n_col,
                                 gridspec_kw=gridspec_kw,
                                 figsize=figsize, dpi=dpi)
        axes = axes.flatten()
        fig.suptitle(main_title)

        for plot_id, trace in enumerate(traces):
            ax = axes[plot_id]
            self.plot_psh(ax, trace,
                          labels,
                          padding,
                          plot_individual=plot_individual,
                          split=split)
            ax.set_title(tittles[plot_id])
            ax.set_xticklabels([])
            ax.set_ylabel('')
            ax.set_xlabel('')

        if show_plot:
            plt.show()

    def make_covariate_psh_figure(self, traces, conditions, padding,
                                  main_title,
                                  tittles,
                                  plot_individual=False,
                                  split=True,
                                  figure_layout=None,
                                  figsize=None,
                                  gridspec_kw=None,
                                  dpi=160,
                                  show_plot=False):
        """
        Plots specified traces for all the traces.
        front_to_tail : how many cycle points to attach from front to tail
        """

        if figure_layout is not None:
            n_rows = figure_layout[0]
            n_col = figure_layout[1]
        else:
            n_rows = len(traces)
            n_col = 1
        if figsize is None:
            figsize = (12, n_rows * 4)

        fig, axes = plt.subplots(n_rows, n_col,
                                 gridspec_kw=gridspec_kw,
                                 figsize=figsize, dpi=dpi)
        axes = axes.flatten()
        fig.suptitle(main_title)

        for plot_id, trace in enumerate(traces):
            ax = axes[plot_id]
            self.plot_covariates_psh(ax, trace,
                                     conditions,
                                     padding,
                                     plot_individual=plot_individual,
                                     split=split)
            ax.set_title(tittles[plot_id])
            ax.set_xticklabels([])
            ax.set_ylabel('')
            ax.set_xlabel('')

        if show_plot:
            plt.show()

    def make_cycle_figure(self, traces,
                          main_title,
                          tittles,
                          plot_individual=True,
                          forward_shift=None,
                          figure_layout=None,
                          figsize=None,
                          gridspec_kw=None,
                          dpi=160,
                          show_plot=False):
        """
        Plots specified traces for all the traces.
        """

        if figure_layout is not None:
            n_rows = figure_layout[0]
            n_col = figure_layout[1]
        else:
            n_rows = len(traces)
            n_col = 1

        if figsize is None:
            figsize = (12, n_rows * 4)

        fig, axes = plt.subplots(n_rows, n_col,
                                 gridspec_kw=gridspec_kw,
                                 figsize=figsize, dpi=dpi)
        axes = axes.flatten()
        fig.suptitle(main_title)

        for plot_id, trace in enumerate(traces):
            ax = axes[plot_id]
            self.plot_cycles(ax, trace,
                             forward_shift=forward_shift,
                             plot_individual=plot_individual)
            ax.set_title(tittles[plot_id])
            ax.set_xticklabels([])
            ax.set_ylabel('')

        if show_plot:
            plt.show()

    def make_avg_act_scat_figure(self,
                                 labels,
                                 main_title,
                                 figure_layout,
                                 cell_numbers=None,
                                 number_cells=False, figsize=(20, 15), dpi=160,
                                 show_plot=False):
        def get_cmap(n, name='hsv'):
            '''
            from : https://stackoverflow.com/questions/14720331/how-to-generate-random-colors-in-matplotlib
            Returns a function that maps each index in 0, 1, ..., n-1 to a distinct
            RGB color; the keyword argument name must be a standard mpl colormap name.'''
            cmap = plt.cm.get_cmap(name, n)
            return [cmap(i_color) for i_color in np.arange(n)]

        fig, axes = plt.subplots(figure_layout[0], figure_layout[1], figsize=figsize, dpi=dpi)

        axes = axes.flatten()
        n_cells = self.traces.shape[1]
        if cell_numbers is None:
            cell_numbers = np.arange(n_cells)
        fig.suptitle(f"{main_title}\nTotal number of cells {n_cells}")

        # get activity at the stimuli
        avg_activity = []
        for label_name in labels:
            label_idx = self.experiment.choose_volumes((self.annotation, label_name))
            signal = self.traces[label_idx, :]
            avg_signal = signal.mean(axis=0)
            avg_activity.append(avg_signal)

        avg_activity = np.vstack(avg_activity)
        cmap = get_cmap(n_cells)

        label_pairs = list(combinations(np.arange(len(labels)), 2))
        # make sure all the points fit
        axis_limits = [min(0, avg_activity.min()) * 100, avg_activity.max() * 1.025 * 100]

        for i_pair, label_pair in enumerate(label_pairs):
            ax = axes[i_pair]
            # in %
            x = avg_activity[label_pair[0]] * 100
            y = avg_activity[label_pair[1]] * 100

            if number_cells:
                ax.scatter(x, y, c=cmap, alpha=0.5)
                for i, txt in enumerate(cell_numbers):
                    ax.annotate(txt, (x[i], y[i]))
            else:
                ax.scatter(x, y, c="grey", alpha=0.5)

            ax.plot(axis_limits, axis_limits, "-k", linewidth=0.5)

            ax.set_xlabel(f'Response to {labels[label_pair[0]]}, avg. dff (%)')
            ax.set_ylabel(f'Response to {labels[label_pair[1]]}, avg. dff (%)')
            ax.set_xlim(axis_limits)
            ax.set_ylim(axis_limits)
            ax.set_aspect(1.0)
        plt.tight_layout(pad=1, w_pad=1, h_pad=1)

        if show_plot:
            plt.show()



import unittest
import tifffile as tif
import numpy as np

from pathlib import Path
import matplotlib.pyplot as plt

import vodex as vx
from numan.utils import *
from tifffile import TiffFile, imread, imwrite


class TestGaussianFilter(unittest.TestCase):
    project_dir = "D:/Code/repos/numan/notebooks/data/v1.x.x/2vs3vs5/processed"
    save_dir = "D:/Code/repos/numan/notebooks/data/v1.x.x/2vs3vs5/test_output"
    experiment = vx.Experiment.load(Path(project_dir, "experiment_raw.db"))

    def test_gaussian_filter_like_Fiji(self):
        """
        MANUAL:
        Load 3 volumes, save them to folder, blur with sigma 1 and save..
        then go to the folder, perform blur with Fiji and compare the two blurs.
        """
        three_first_volumes = self.experiment.load_volumes([0, 1, 3])
        print(type(three_first_volumes))
        print(type(three_first_volumes[0, 0, 0, 0]))

        # write image
        imwrite(f'{self.save_dir}/three_volumes.tif',
                three_first_volumes.astype(np.uint16), shape=(3, 52, 468, 500),
                metadata={'axes': 'TZYX'}, imagej=True)

        sigma = 1
        three_first_volumes = gaussian_filter(three_first_volumes, sigma)
        # write blurred image
        imwrite(f'{self.save_dir}/three_volumes_blurred.tif',
                three_first_volumes.astype(np.uint16), shape=(3, 52, 468, 500),
                metadata={'axes': 'TZYX'}, imagej=True)

class TestGetDffInSteps(unittest.TestCase):
    array = np.ones((10, 2, 2))
    array[:, 0, 0] = [2, 2, 4, 2, 2, 3, 3, 6, 3, 3]
    step_size = 5
    baseline_volumes = [0, 1]

    def test_get_dff_in_steps(self):
        dff, baseline = get_dff_in_steps(self.array, self.step_size, self.baseline_volumes)

        true_baseline = np.ones((10, 2, 2))
        true_baseline[:, 0, 0] = [2, 2, 2, 2, 2, 3, 3, 3, 3, 3]

        true_dff = np.zeros((10, 2, 2))
        true_dff[:, 0, 0] = [0, 0, 1, 0, 0, 0, 0, 1, 0, 0]

        assert np.array_equal(baseline, true_baseline)
        assert np.array_equal(dff, true_dff)

if __name__ == '__main__':
    unittest.main()


"""
Helps set up the project structure.
"""
import os


class Project:
    def __init__(self, project_folder):
        self.main_folder = project_folder

    def create(self, folder_in_project):
        path = os.path.join(self.main_folder, folder_in_project)
        assert not os.path.isdir(path), f'the directory {path} already exists in the project,' \
                                        ' did you mean to process another project?'
        os.makedirs(path)

    def activate(self, folder_in_project):
        path = os.path.join(self.main_folder, folder_in_project)
        assert os.path.isdir(path), f'Can not activate the folder {path}, the directory does not exists.'
        os.chdir(path)

    def check_exists(self, folder_in_project):
        path = os.path.join(self.main_folder, folder_in_project)
        assert os.path.isdir(path), f'Directory {path} does not exists. ' \
                                    f'Verify that this is the right project folder ' \
                                    f'and that you have run the previous analysis steps.'



import urllib3
import json
import traceback
import base64

# read more https://api.slack.com/methods/chat.postMessage
# and here is how to listen to messages: https://slack.dev/bolt-python/concepts !!!

# this is not very smart...I know... please don't post anything to our channel :D
webhook_url = base64.b64decode(
    'aHR0cHM6Ly9ob29rcy5zbGFjay5jb20vc2VydmljZXMvVFBFRkw3QjZDL' \
    '0IwM1UyNllDNDhNL0hhdHAxYUFyVE9JNElsbTA0VWRLVzJIbg=='.encode('ascii')).decode('ascii')

user_id = {"anna": "UPGMQ34BG", "peter": "UPE21JHRA"}


# Send Slack notification based on the given message
def slack_notification(message, tag_users=None):
    if tag_users is not None:
        tags = ""
        for user in tag_users:
            tags = tags + f'<@{user_id[user]}>'
        message = f'{tags}\n {message}'

    try:
        slack_message = {'text': message}

        http = urllib3.PoolManager()
        response = http.request('POST',
                                webhook_url,
                                body=json.dumps(slack_message),
                                headers={'Content-Type': 'application/json'},
                                retries=False)
    except:
        traceback.print_exc()

    return True


if __name__ == "__main__":
    slack_notification(f'go to sleep now ?',
                       tag_users=["anna"])



from tifffile import TiffFile, imread, imsave
import numpy as np
import json
import os

import matplotlib.pyplot as plt
import warnings
from tqdm import tqdm
import scipy.ndimage as nd
import pandas as pd
import PyPDF2


def extract_windows(array, window_size):
    """
    breaks ND array into windows along 0 dimention, starting from the begining and till the end.
    """
    start = 0
    num_windows = array.shape[0] - window_size + 1

    sub_windows = (
            start +
            np.expand_dims(np.arange(window_size), 0) +
            np.expand_dims(np.arange(num_windows), 0).T
    )
    return array[sub_windows]


def sort_by_len0(zip_to_sort):
    """
    Sorts the zip based on the length and alphabet of the first element in zip_to_sort
    """
    # sorts a list based on the length of the first element in zip
    sorted_zip = sorted(zip_to_sort, key=lambda x: (len(x[0])), reverse=True)
    return sorted_zip


def get_baseline(array, window_size, percentile):
    """
    returns average baseline of an ND array along the 0 dimention.
    window_size must be an odd number.

    The baseline for the first and the last window//2 elements is the same :
    the first or the last value calculated with the full window
    """
    mean_signal = np.percentile(extract_windows(array, window_size), percentile, axis=1)
    # construct the baseline
    start = window_size // 2
    end = len(array) - window_size // 2
    # add the beginning and end to baseline
    baseline = np.zeros(array.shape)
    baseline[0:start] = mean_signal[0]
    baseline[start:end] = mean_signal
    baseline[end:] = mean_signal[-1]

    return baseline, start, end

def check_baseline_not_zero(baseline):
    """
    Checks if baseline is zero, if so, sets it to 10^(-6)
    """
    b_zero = baseline == 0
    if np.sum(b_zero) > 0:
        warnings.warn(f"{np.sum(b_zero)} baseline values are zero.\n"
                      f"Setting these values to 10^(-6) to avoid dividing by zero.")
        baseline[b_zero] = 10 ^ (-6)

    return baseline


def get_dff(array, window_size):
    """
    subtracts 8th percentile baseline from an ND array along the 0 dimention.
    window_size must be an odd number.

    The baseline for the first and the last window//2 elements is the same :
    the first or the last value calculated with the full window
    """
    percentile = 8  # 8th percentile
    baseline, start, end = get_baseline(array, window_size, percentile)

    # not to divide by zero:
    baseline = check_baseline_not_zero(baseline)

    return (array - baseline) / baseline, start, end

def get_dff_in_steps(array, step_size, baseline_volumes):
    """
    Subtracts baseline calculated for each step from an ND array along the 0 dimention and divides by baseline.

    Args:
        array (NumpyArray): 4D array (TZYX) to calculate dff for
        step_size (int): number of volumes to correct for a single baseline
        baseline_volumes (List[int]): number of volumes to average for baseline inside each step, 
            for example [0,1,2] will  average the first 3 volumes inside each step to calculate baseline.
    """
    #split array into steps of step_size volumes each
    n_crops = array.shape[0]//step_size

    #calculate baseline for each step
    baseline = np.zeros(array.shape)
    for i in range(n_crops):
        from_idx = i*step_size
        to_idx = (i+1)*step_size
        baseline[from_idx:to_idx] = np.mean(array[from_idx:to_idx][baseline_volumes], axis=0)
    # not to divide by zero:
    baseline = check_baseline_not_zero(baseline)

    return (array - baseline) / baseline, baseline



    



def get_t_score(movie1, movie2, absolute=True):
    """
    Returns absolute t-score image ( 2D or 3D, depending on input),
    for t-score calculations see for example :
    https://www.jmp.com/en_us/statistics-knowledge-portal/t-test/two-sample-t-test.html
    """
    # TODO : check input dimensions

    avg1 = np.mean(movie1, axis=0)
    avg2 = np.mean(movie2, axis=0)

    var1 = np.var(movie1, axis=0)
    var2 = np.var(movie2, axis=0)
    n1 = movie1.shape[0]
    n2 = movie2.shape[0]

    std = np.sqrt(((n1 - 1) * var1 + (n2 - 1) * var2) / (n1 + n2 - 2))

    # not to divide by zero:
    std_zero = std == 0
    if np.sum(std_zero) > 0:
        warnings.warn(f"{np.sum(std_zero)} std values are zero.\n"
                      f"Setting these values to 10^(-6) to avoid dividing by zero.")
        std[std_zero] = 10 ^ (-6)

    if absolute:
        t_score = np.absolute(avg1 - avg2) / std
    else:
        t_score = (avg1 - avg2) / std

    return t_score


def get_diff(movie1, movie2, absolute=False):
    """
    Returns absolute difference image ( 2D or 3D, depending on input).
    per pixel: calculates the mean image for each movie, subtracts and takes the absolute value.
    """
    # TODO : check input dimensions

    avg1 = np.mean(movie1, axis=0)
    avg2 = np.mean(movie2, axis=0)

    if absolute:
        diff = np.absolute(avg1 - avg2)
    else:
        diff = (avg1 - avg2)

    return diff


def gaussian_filter(movie, sigma):
    """
    Wrapper for scipy.ndimage gaussian_filter to use with 4D data.
    movie : 4D movie: TZYX
    sigma : gaussian filter parameters, ZYX order , in pixels
    (from scipy : scalar or sequence of scalars Standard deviation for Gaussian kernel. The
    standard deviations of the Gaussian filter are given for each axis as a sequence, or as a single number,
    in which case it is equal for all axes.)
    """
    for iv, volume in enumerate(movie):
        movie[iv] = nd.gaussian_filter(volume, sigma)
    return movie


def plot_errorbar(ax, mean, e, x=None, color='r', edge_color=None):
    if x is None:
        x = np.arange(len(mean))
    ax.errorbar(x, mean, yerr=e, fmt='o', color=color, markeredgecolor=edge_color)
    ax.plot(x, mean, color=color)


def get_ax_limits(cycled, mean, e, plot_individual):
    """
    Figures out the tight x and y axis limits
    """
    if plot_individual:
        ymin = np.min(cycled)
        ymax = np.max(cycled)
    else:
        ymin = np.min(mean - e[1, :])
        ymax = np.max(mean + e[0, :])
    xmin = -0.5
    xmax = cycled.shape[1] - 0.5

    return xmin, xmax, ymin, ymax



"""
Classes for Numerosity analysis.
"""
from tifffile import TiffFile, imread, imwrite
from pathlib import Path
import numpy as np
import scipy as sp
import json
import os

import matplotlib.pyplot as plt
import warnings
from tqdm import tqdm
import pandas as pd
import PyPDF2

# try:
import ants
# except ImportError:
#     print("ants not installed, some registration will not work")

from .utils import *


class Spot:
    """
    This is a class for a n individual segmented spot.
    """

    def __init__(self, center, diameter, resolution=None, units='pix'):
        """
        Parameters
        ----------
        center : list or numpy array, center in zyx order
        diameter : list or numpy array or int, diameter in zyx order. If int, then it is a sphere.
        resolution : list or numpy array, resolution in xyz order.
        Defaults to [1,1,1] if not specified and only pixels are used.
        info : any extra info about the spots
        units : what units the center and diameter are using , physical 'phs', or pixels , 'pix'.
        When using physical units you must provide the resolution.
        """
        self.diameter = {}
        self.center = {}

        center = np.array(center)
        if isinstance(diameter, int):
            diameter = np.array([1, 1, 1]) * diameter
        else:
            diameter = np.array(diameter)

        if resolution is not None:
            self.resolution = np.array(resolution)
            if units == 'phs':

                self.center['phs'] = center
                self.diameter['phs'] = diameter

                self.center['pix'] = np.round(center / self.resolution)
                self.diameter['pix'] = np.round(diameter / self.resolution)
                # in case we get a 0 diameter with all the rounding
                self.diameter['pix'][self.diameter['pix'] == 0] = 1

            elif units == 'pix':
                self.center['phs'] = center * self.resolution
                self.diameter['phs'] = diameter * self.resolution

                self.center['pix'] = center
                self.diameter['pix'] = diameter
            else:
                raise AssertionError("Units can be 'pix' or 'phs' only ")

        if resolution is None:
            self.resolution = None
            if units == 'phs':
                raise AssertionError('when using physical units, you must provide the resolution')

            elif units == 'pix':

                self.center['phs'] = None
                self.diameter['phs'] = None

                self.center['pix'] = center
                self.diameter['pix'] = diameter
            else:
                raise AssertionError("Units can be 'pix' or 'phs' only ")

    def __str__(self):
        return f"Spot at {self.center['phs']} phs, {self.center['pix']} pix\n" \
               f"diameter {self.diameter['phs']} phs, {self.diameter['pix']} pix\n" \
               f"resolution {self.resolution}. Everything in ZYX order."

    def __repr__(self):
        return self.__str__()

    def mask_at_zero(self, diameter=None, units='pix'):
        """
        diameter : list or numpy array or int, diameter in zyx order. If int, then it is a sphere.
                    If not None, diameter will be used to create a mask.
                    If None, then spots diameter will be used and units will be ignored.
        units: what units is the diameter in : 'pix' or 'phs'
        """

        if diameter is None:
            d = self.diameter['pix']
        else:
            if isinstance(diameter, int):
                diameter = np.array([1, 1, 1]) * diameter
            else:
                diameter = np.array(diameter)
            if units == 'phs':
                diameter = np.round(diameter / self.resolution)
            d = diameter

        r = d / 2  # radius
        # make sure diameter is odd: round to the next odd number
        d = ((d // 2) * 2 + 1).astype(int)
        # center at zero
        c = (d // 2).astype(int)
        z = np.arange(d[0]) - c[0]
        y = np.arange(d[1]) - c[1]
        x = np.arange(d[2]) - c[2]

        # find what pixels are inside the ellipsoid
        zz, yy, xx = np.meshgrid(z, y, x, indexing='ij')
        inside_ellipsoid = np.sqrt((zz / r[0]) ** 2 + (yy / r[1]) ** 2 + (xx / r[2]) ** 2) <= 1

        # get a 3D mask
        mask = np.zeros((d[0], d[1], d[2]))
        mask[inside_ellipsoid] = 1

        # get list of nonzero elements , centered at zero
        i, j, k = np.where(mask > 0)
        centered_idx = np.c_[i, j, k] - c

        return mask, centered_idx

    def create_mask(self, volumes, diameter=None, units='pix'):
        """
        Creates a binary mask that can be applied to the volume.
        volumes : image as a sequence of volumes in shape TZYX or one volume in ZYX
        diameter : list or numpy array or int, diameter in zyx order. If int, then it is a sphere.
                    If not None, diameter will be used to create a mask.
                    If None, then spots diameter will be used and units will be ignored.
        units: what units is the diameter in : 'pix' or 'phs'
        """
        # create mask for a single volume
        if len(volumes.shape) == 4:
            t, zmax, ymax, xmax = volumes.shape
        elif len(volumes.shape) == 3:
            zmax, ymax, xmax = volumes.shape
        else:
            raise AssertionError("volumes should be 4D tzyx or 3D zyx")
        mask = np.zeros((zmax, ymax, xmax))

        # get list of nonzero elements , centered at zero
        _, idx = self.mask_at_zero(diameter=diameter, units=units)

        # shift to center
        shift = self.center['pix']
        # because of the rounding when going from phs to pix, some extreme spots can be outside the image
        # so make sure they are inside
        border = np.array([zmax - 1, ymax - 1, xmax - 1])
        shift = np.min(np.c_[shift, border], axis=1)

        # get the indices
        idx = idx + shift.astype(int)

        # remove the ones outside the boundary
        idx = idx[np.all(idx >= 0, axis=1), :]
        is_inside = np.logical_and(idx[:, 0] < zmax, np.logical_and(idx[:, 1] < ymax, idx[:, 2] < xmax))
        idx = idx[is_inside, :]

        mask[idx[:, 0], idx[:, 1], idx[:, 2]] = 1

        if np.sum(mask) == 0:
            warnings.warn(f"The whole spot centered at {self.center['pix']} pix ( {self.center['phs']} phs), with "
                          f"diameter {self.diameter['pix']} pix ({self.diameter['phs']} phs) is "
                          f"outside of the volume.")

        return mask, idx

    @staticmethod
    def plot_mask(mask, figsize=(8, 6), dpi=160):
        n_slices = mask.shape[0]
        if n_slices > 1:
            f, ax = plt.subplots(1, mask.shape[0], figsize=figsize, dpi=dpi)
            for islice, iax in enumerate(ax):
                iax.imshow(mask[islice, :, :], vmin=0, vmax=1)
                iax.set_axis_off()
                iax.set_title(f"z {islice}")
                iax.grid(which='minor', color='w', linestyle='-', linewidth=2)
        else:
            plt.imshow(mask[0, :, :], vmin=0, vmax=1)


class Signals:
    def __init__(self, traces, traces_type="raw", fps=1):
        """
        traces : matrix TxN ( time x number of signals )
        traces_type : a 3 letter code for what kind of signals are these : "raw", "dff", "zsc" ( z  - score )
        """
        self.traces = np.array(traces)
        self.T, self.N = self.traces.shape

        self.traces_type = traces_type
        self.fps = fps

    def change_fs(self, fps):
        self.fps = fps

    def as_dff(self, window_size):
        """
        Returns dff of the design_matrix, the beginning of the proper measurement and the end
        ( such measurement that used the whole window for definition of baseline )
        """
        assert self.traces_type == "raw", f"Can't apply dff: " \
                                          f"the signals have already been processed, these are {self.traces_type} signals"
        dff, start, end = get_dff(self.traces, window_size)

        return Signals(dff, traces_type="dff")

    def hp_filter(self, cutoff):
        """
        Returns high-pass filtered signals from the design_matrix
        """
        assert self.traces_type == "raw", f"Can't apply dff: " \
                                          f"the signals have already been processed, these are {self.traces_type} signals"

        def butter_highpass(cutoff, fps, order=5):
            # calculate the numerator and denominator coefficient vectors of the filter
            nyq = 0.5 * fps
            normal_cutoff = cutoff / nyq
            b, a = signal.butter(order, normal_cutoff, btype="high", analog=False)
            return b, a

        def butter_highpass_filter(data, cutoff, fps, order=5):
            # return the filtered signal
            b, a = butter_highpass(cutoff, fps, order=order)
            y = signal.filtfilt(b, a, data, axis=0)
            return y

        filtered_traces = butter_highpass_filter(self.traces, cutoff, self.fps)

        return Signals(filtered_traces, traces_type="hp filtered", fps=self.fps)

    def as_zscore(self, cutoff, window_size):
        """
        Returns z-score of the design_matrix
        Uses high-pass filter for z-score calculation
        """
        assert self.traces_type == "raw", f"Can't apply dff: " \
                                          f"the signals have already been processed, these are {self.traces_type} signals"
        # get baseline : as median
        percentile = 50
        baseline, start, end = get_baseline(self.traces, window_size, percentile)
        # get standart deviation from high-pass filtred data
        filtered_signals = self.hp_filter(cutoff)
        filtered_sd = np.std(filtered_signals.traces, axis=0)
        # calcualte z-score
        zscore_signals = (self.traces - baseline) / filtered_sd

        return Signals(zscore_signals, traces_type="zscore", dff=self.dff)

    @classmethod
    def from_spots(cls, spots, volumes=None, experiment=None, batch_size=None, movie=None,
                   verbose=False, traces_type="raw"):
        """
        Extracts signals from volumes for each spot in spots.
        spots : list[Spot]
        volumes : volume idx (list[int]) to load and extract signals from, or "all"
        movie : volume sequence (tzyx) to extract signals from.
        """

        assert (movie is not None) or (volumes is not None and experiment is not None), \
            "Provide a movie or volume IDs and an experiment "

        if volumes == "all":
            # only use full volumes
            volumes = experiment.list_volumes()
            volumes = volumes[volumes >= 0]

        def fill_signal(matrix, movie, t_start, t_end):
            for isp, spot in enumerate(spots):
                _, idx = spot.create_mask(movie)
                # gets the pixels in the spot mask
                avg_signal = np.mean(movie[:, idx[:, 0], idx[:, 1], idx[:, 2]], axis=1)
                matrix[t_start:t_end, isp] = avg_signal
            return matrix

        N = len(spots)

        design_matrix = None
        # if volumes are given explicitly as a movie
        if movie is not None:
            T = movie.shape[0]
            design_matrix = np.zeros((T, N))
            design_matrix = fill_signal(design_matrix, movie, 0, T)

        # if you need to load volumes from disk
        elif volumes is not None and experiment is not None:
            T = len(volumes)
            design_matrix = np.zeros((T, N))
            # split volumes into batches of certain size
            if batch_size is not None:
                btcs = np.array_split(volumes, np.ceil(T / batch_size))
            else:
                btcs = volumes

            t_start = 0
            for batch in btcs:
                t_end = t_start + len(batch)
                if verbose:
                    print(f"Batch {batch}, t_start {t_start}, t_end {t_end}")
                # load a batch from disk
                movie = experiment.load_volumes(batch, verbose=True)
                # extract signals for all the spots from the loaded chunk
                design_matrix = fill_signal(design_matrix, movie, t_start, t_end)

                t_start = t_end

        return cls(design_matrix, traces_type=traces_type)

    def get_looped(self, trace, experiment, time_points=None, cycles=None, error_type="prc"):
        """
        Returns signals looped per cycle
        time_points: time points of the cycle. If you only need certain time-points from the cycle ( in volumes )
        if you need to stack timepoints for each cycle, for example: [[0,1,2],[7,8,9]], pass the time_points as 2D list.
        """

        n_cycles = experiment.full_cycles
        # cycle length in volumes
        cycle_length = experiment.cycle.full_length / experiment.volume_manager.fpv
        assert cycle_length.is_integer(), "Need more info to determine the cycle..." \
                                          "by which I mean that plot_looped function needs more code :), " \
                                          "Sorry. Hint : use time_points "
        assert self.T == (n_cycles * cycle_length), "Need more info to determine the cycle..." \
                                                    "by which I mean that plot_looped function needs more code :), " \
                                                    "Sorry. Hint : use time_points "
        cycled = self.traces[:, trace].reshape((n_cycles, int(cycle_length)))

        # grab only the desired cycles
        if cycles is not None:
            cycles = np.array(cycles)
            cycled = cycled[cycles, :]

        # crop out the necessary cycle part,
        # note: you can reorder the cycle as well ( to add points from the beginning  to the end of the cycle)
        # you can also overlay some cycle time intervals, useful for psh
        if time_points is not None:
            time_points = np.array(time_points)
            time_shape = time_points.shape
            assert len(time_shape) < 3, "time_shape should be 1D or 2D"
            cycled = cycled[:, time_points.flatten()]
            # if you want to overlay cycle intervals:
            if len(time_shape) == 2:
                cycled = cycled.reshape(-1, time_shape[1])

        mean = np.mean(cycled, axis=0)

        if error_type == "prc":
            # error bars : 5 to 95 th percentile around the median
            e = np.r_[np.expand_dims(mean - np.percentile(cycled, 5, axis=0), axis=0),
            np.expand_dims(np.percentile(cycled, 95, axis=0) - mean, axis=0)]
        elif error_type == "sem":
            # error bars : sem around hte mean
            sem = np.std(cycled, axis=0, ddof=1) / np.sqrt(cycled.shape[0])
            e = np.r_[np.expand_dims(sem, axis=0),
            np.expand_dims(sem, axis=0)]
        else:
            e = None

        return cycled, mean, e

    def to_csv(self, filename):
        """
        saves signals to csv.
        """
        df = pd.DataFrame(self.traces, columns=[f"cell {icell}" for icell in np.arange(self.N)])
        df.to_csv(filename)


class Spots:

    def __init__(self, spots, groups=None, signals=None):
        """
        Parameters
        ----------
        spots : list[Spot]
        groups : dictionary of type {"group_name" : list[bool]} that assigns cells to various groups
        signals : Signals for the corresponding spots
        """

        self.spots = spots
        self.num_spots = len(self.spots)
        self.signals = signals
        # TODO : have signals as dict with " raw" , " dff" etc
        self.groups = None
        if groups is not None:
            self.add_groups(groups)

    def __str__(self):

        groups = "No"
        if self.groups is not None:
            groups = self.groups.keys()

        signals = "not loaded"
        if self.signals is not None:
            signals = "loaded"

        return f"{self.num_spots} spots\n " \
               f"{groups} groups\n" \
               f"Signals {signals}"

    def __repr__(self):
        return self.__str__()

    def add_groups(self, groups, rewrite=False):
        """
        groups: dict with boolean arrays to say which cells belong.
        """
        if not rewrite and self.groups is not None:
            for group in groups:
                assert group not in self.groups, f"The group {group} already exists," \
                                                 f" use rewrite to overwrite it," \
                                                 f" or use another group name"
        if self.groups is None:
            self.groups = {}
        for group in groups:
            # TODO : add protection from rewriting
            self.groups[group] = np.array(groups[group])

    def list_groups(self):
        if self.groups is not None:
            for key in self.groups.keys():
                print(key)

    def _get_centers(self, units='phs'):
        """
        returns a list of lists
        """
        centers = []
        for spot in self.spots:
            centers.append(spot.center[units].tolist())
        return centers

    def _get_diameters(self, units='phs'):
        diameters = []
        for spot in self.spots:
            diameters.append(spot.diameter[units].tolist())
        return diameters

    def _get_resolutions(self):
        resolutions = []
        for spot in self.spots:
            resolutions.append(spot.resolution.tolist())
        return resolutions

    def to_json(self, filename, units='phs'):
        """
        Transform Spots object into json format and save as a file.
        """

        j_dict = {
            "resolutions": self._get_resolutions(),
            "diameters": self._get_diameters(units=units),
            "centers": self._get_centers(units=units),
            "units": units,
            "groups": None}

        if self.groups is not None:
            j_dict["groups"] = {}
            for group in self.groups:
                j_dict["groups"][group] = np.array(self.groups[group], dtype=bool).tolist()

        if self.signals is not None:
            j_dict["signals"] = {"traces": self.signals.traces.tolist(),
                                 "traces_type": self.signals.traces_type}
        else:
            j_dict["signals"] = None

        j = json.dumps(j_dict)

        with open(filename, 'w') as json_file:
            json_file.write(j)

    @classmethod
    def from_json(cls, filename):
        """
        Load Spots object from json file.
        """
        # create an object for the class to return
        with open(filename) as json_file:
            j = json.load(json_file)

        spots = []
        units = j["units"]
        for center, diameter, resolution in zip(j["centers"], j["diameters"], j["resolutions"]):
            spots.append(Spot(center, diameter, resolution=resolution, units=units))

        groups = j["groups"]

        if j["signals"] is not None:
            signals = Signals(j["signals"]["traces"], traces_type=j["signals"]["traces_type"])
        else:
            signals = None

        return cls(spots, groups=groups, signals=signals)

    @classmethod
    def from_imaris(cls, points_file, diameter_file, resolution=None, units='phs'):
        """
        Load Spots using imaris csvs:
        Creates list of spots from the output of Imaris segmentation.

        position_file: *_Position.csv , file containing segmented centers
        diam_file: *_Diameter.csv , file containing diameters
        """
        p_df = pd.read_csv(points_file, skiprows=[0, 1, 2])
        d_df = pd.read_csv(diameter_file, skiprows=[0, 1, 2])
        centers = p_df[['Position Z', 'Position Y', 'Position X']].to_numpy()
        diams = d_df[['Diameter Z', 'Diameter Y', 'Diameter X']].to_numpy()
        spots = []
        for center, diam in zip(centers, diams):
            spots.append(Spot(center, diam, resolution=resolution, units=units))

        return cls(spots, groups=None, signals=None)

    def get_signals(self, volumes=None, experiment=None, batch_size=None, movie=None, traces_type="raw", reload=False):
        assert reload or self.signals is None, "Spots already have signals loaded, reload? "

        self.signals = Signals.from_spots(self.spots, volumes=volumes, experiment=experiment,
                                          batch_size=batch_size, movie=movie, traces_type=traces_type)

    def get_group_mask(self, group, mask_shape, diameter=None, units='pix'):
        """
        Create a 3D volume that only shows the spots that belong to the particular group
        group : list[bool], the length of spots
        mask_shape : the shape of the 3D volume of the mask
        diameter : list or numpy array or int, diameter in zyx order. If int, then it is a sphere.
                    If not None, diameter will be used to create a mask.
                    If None, then spots diameter will be used and units will be ignored.
        units: what units is the diameter in : 'pix' or 'phs'
        """
        mask = np.zeros(mask_shape)
        for ispot in tqdm(np.where(group)[0]):
            imask, _ = self.spots[ispot].create_mask(mask, diameter=diameter, units=units)
            mask = mask + imask
        return mask

    def get_group_signals(self, group):
        """
        Returns signals for a particular group only
        """
        traces = self.signals.traces[:, group]
        return Signals(traces, traces_type=self.signals.traces_type)

    def get_group_idx(self, group):
        """
        Returns indices for a particular group only
        """
        idx = np.arange(self.num_spots)
        idx = idx[group]
        return idx

    def get_group_centers(self, group, units='pix'):
        """
        Returns centers for a particular group only
        units: the units of the center to return
        """
        centers = []
        for ispot, spot in enumerate(self.spots):
            if group[ispot]:
                centers.append(spot.center[units])

        return np.array(centers)

    def get_group_info(self, group_list, group=None):
        """
        Returns a string with the titles of the groups from group_list, where each spot is a member.
        group = the T/F list ... If group is not None: only returns it for the group.
        """
        group_info = []
        for ispot, spot in enumerate(self.spots):
            groups = ""
            for group_name in group_list:
                if self.groups[group_name][ispot]:
                    groups = groups + group_name + "; "
            group_info.append(groups)
        group_info = np.array(group_info)
        if group is not None:
            group_info = group_info[group]
        return group_info


class SignalAnalyzer:

    def __init__(self, signals):
        self.signals = signals

    @staticmethod
    def bootstrap_distribution_of_difference_of_means(data, split, nbootstrap=10000):
        """
        Creates a bootstrap distribution of the absolute difference of the means for the two groups

        data : list of data to split into the two groups and calculate the difference of the mean
        split : [n_group1,n_group2]
        nbootstrap : number of bootstrap repetitions
        """
        N = data.shape[0]
        if np.sum(split) != N:
            raise AssertionError("sum of split should equal the number of data points")
        delta = []
        for ib in np.arange(nbootstrap):
            group1 = np.random.choice(N, size=split[0], replace=True)
            group2 = np.arange(N)[~group1]
            delta.append(np.mean(data[group1]) - np.mean(data[group2]))
        return delta

    @staticmethod
    def get_p_of_difference_of_means(data1, data2, nbootstrap=10000):

        split = [len(data1), len(data2)]
        merged_data = np.r_[data1, data2]

        delta = np.abs(np.mean(data1) - np.mean(data2))
        # get distribution of the absolute diff of the means
        delta_dist = SignalAnalyzer.bootstrap_distribution_of_difference_of_means(merged_data, split=split,
                                                                                  nbootstrap=nbootstrap)
        delta_dist = np.abs(delta_dist)

        n_samples = len(delta_dist)
        n_extreme = np.sum(delta_dist >= delta)
        p = n_extreme / n_samples
        return p

    def get_p_list_of_difference_of_means(self, group1, group2, nbootstrap=10000):
        """
        group1 : data points ( time points, in volumes ) to assign to group 1
        group2 : data points ( time points, in volumes ) to assign to group 2
        """
        data1 = self.signals.traces[group1, :]
        data2 = self.signals.traces[group2, :]
        T1, nspots = data1.shape
        T2 = data1.shape[0]
        # print(f" Data points in group 1 : {T1},in group 1 : {T2}.\nNumber of spots : {nspots}.")

        p = []
        for ispot in tqdm(np.arange(nspots)):
            p.append(self.get_p_of_difference_of_means(data1[:, ispot], data2[:, ispot], nbootstrap=nbootstrap))
        return p


class Volumes:
    """
    Collection of volumes and functions that perform operations on them.
    """

    def __init__(self, volumes, resolution=[1, 1, 1]):
        self.volumes = volumes
        self.resolution = resolution
        self.shape = volumes.shape

    def average_volumes(self):
        return Volumes(np.mean(self.volumes, axis=0))

    def voxelise(self):
        pass

    def get_dff(self, window_size):
        """
        Returns dff of the volumes (assuming 0 dimension is time),
        the beginning of the proper measurement and the end
        ( such measurement that used the whole window for definition of baseline )
        """
        dff, start, end = get_dff(self.volumes, window_size)
        return dff, start, end


class Preprocess:
    """
    Collection of methods to perform preprocessing of the raw data in experiment.
    """

    def __init__(self, experiment):
        self.experiment = experiment

    def remove_offset(self, save_dir, batch_size, volumes = None, offset_img=None, offset_file=None, verbose=False):
        """
        Denoises the raw movie by subtracting the offset image.
        The offset image can be prepared using coscmos package.
        Will only process full volumes.
        """
        assert offset_img is not None or offset_file is not None, \
            "Either offset_img or offset_file should be provided"
        assert not (offset_img is not None and offset_file is not None), \
            "Only one of offset_img or offset_file should be provided"
        if offset_img is None:
            offset_img = imread(offset_file)

        # Split the volumes into chunks of the defined size that will be loaded into RAM at once
        # if volumes are not provided, will use all the full volumes
        # if volumes are provided, will only process the requested volumes
        chunks = self.experiment.batch_volumes(batch_size, volumes=volumes, full_only=True)
        
        if verbose:
            tot_volumes = 0
            n_requested_volumes = self.experiment.n_full_volumes
            if volumes is not None:
                n_requested_volumes = len(volumes)

        for ichunk, chunk in enumerate(chunks):
            data = self.experiment.load_volumes(chunk, verbose=False)
            data = data - offset_img
            nt, z, y, x = data.shape

            # make sure the output is positive:
            min_value = data.min()
            if min_value < 0:
                warnings.warn(f"Some image values belov zero after subtracting the offset image. "
                              f"Will be set to zero.")
                data[data < 0] = 0

            # write image
            imwrite(f'{save_dir}/removed_offset_movie_{ichunk:04d}.tif',
                    data.astype(np.int16), shape=(nt, z, y, x),
                    metadata={'axes': 'TZYX'}, imagej=True)

            if verbose:
                tot_volumes = tot_volumes + nt
                print(f"written volumes : {tot_volumes}, out of {n_requested_volumes} full volumes")

    def drift_correct_naive(self, save_dir, batch_size, spacing_xyz, template = 0, 
                            registration_type = "Rigid", volumes = None, verbose=False, 
                            ants_kwargs = {}):
        """
        Performs drift correction using ANTs library for registration. 
        template can be an int (a frame to register to), or a 3D numpy array.
        If template is not provided, the first volume will be used as a template.
        if volumes is not None, then only the volumes in the list will be registered.
        For more info, see ANTs documentation: https://antspy.readthedocs.io/en/latest/registration.html

        Args:
            save_dir (str): directory to save the drift corrected movie.
            batch_size (int): number of volumes to load at once when processing batches. Each batch is saved as a separate file.
            spacing_xyz (list): spacing in xyz order
            template (int or 3D numpy array): template to register to
            registration_type (str): type of registration to perform, see ANTs documentation for available options.
            volumes (list): list of volumes to register. If None, all the full volumes will be registered.
            verbose (bool): whether to print the volumes that have been processed so far on the screen.
            ants_kwargs (dict): parameters for the registration. If empty, default parameters will be used. See ANTs documentation for available options.
                mask and moving_mask can be provided as paths to the masks, they will be converted to ants images.
                Note: (see https://stackoverflow.com/questions/334655/passing-a-dictionary-to-a-function-as-keyword-parameters)
                 1. ants.registration can have parameters that are not included in the dictionary, just specify what you want, the rest will be default. 
                 2. You can not override an ants.registration function parameter that is already in the dictionary, 
                 so don't try to include keywords fixed, moving, type_of_transform, verbose and outprefix in the dictionary.
                 3. The dictionary can not have values that aren't in the ants.registration function.
        """
        def _register(moving_image, fixed_image, spacing, registration_type, outprefix = None):

            # if fixed image in not ants image, convert it
            if not isinstance(fixed_image, ants.core.ants_image.ANTsImage):
                print("Converting fixed image to ants image")
                fixed_image = ants.from_numpy(np.transpose(fixed_image.astype(float), axes=[2,1,0]), 
                                    spacing=spacing) # spacing in xyz order
                
            moving_image = ants.from_numpy(np.transpose(moving_image.astype(float), axes=[2,1,0]), 
                                spacing=spacing)
            
            # run the registration and save verbose to file
            rr = ants.registration(
                                    fixed=fixed_image,
                                    moving=moving_image,
                                    type_of_transform=registration_type, 
                                    verbose  = False,
                                    outprefix = outprefix,
                                    **ants_kwargs
                                )
            return rr['warpedmovout'].numpy().astype(np.int16).transpose((2,1,0))
        
        def _register_batch(batch, template, spacing, registration_type, outprefix = None):
            corrected = []
            for volume in batch:
                # load the volume
                volume_img = self.experiment.load_volumes([volume], verbose=False)[0]
                # register
                volume_img = _register(volume_img, template, spacing, registration_type, 
                                       outprefix = outprefix)
                corrected.append(volume_img)
            corrected = np.array(corrected)
            return corrected

        # prepare list of volumes to register
        n_full_volumes = self.experiment.n_full_volumes
        if volumes is not None:
            volumes = np.array(volumes)
            assert volumes.max() < n_full_volumes,\
                  f"Some volumes are outside of the experiment with {n_full_volumes} full volumes" 
            assert volumes.min() >= 0, "volumes must be positive integers"
            n_reguested_volumes = len(volumes)
        else:
            volumes = np.arange(n_full_volumes)
            n_reguested_volumes = n_full_volumes

        # if template is not an array, get template volume
        if isinstance(template, int):
            assert template < n_full_volumes, f"Template frame {template} is outside of the experiment with {n_full_volumes} full volumes"
            assert template >= 0, "Template frame must be a positive integer"
            template = self.experiment.load_volumes([template], verbose=False)[0]
        # turn template into ants image
        template = ants.from_numpy(np.transpose(template.astype(float), axes=[2,1,0]), 
                                    spacing=spacing_xyz) # spacing in xyz order
        
        # if mask and moving_mask are in ants_kwargs, convert them to ants images
        # and if they are string or pthlib.Path, load them as numpy arrays first
        if 'mask' in ants_kwargs:
            if isinstance(ants_kwargs['mask'], str) or isinstance(ants_kwargs['mask'], Path):
                ants_kwargs['mask'] = ants.from_numpy(np.transpose(imread(ants_kwargs['mask']).astype(float), axes=[2,1,0]), 
                                    spacing=spacing_xyz)
            elif isinstance(ants_kwargs['mask'], np.ndarray):
                ants_kwargs['mask'] = ants.from_numpy(np.transpose(ants_kwargs['mask'].astype(float), axes=[2,1,0]), 
                                    spacing=spacing_xyz)

        if 'moving_mask' in ants_kwargs:
            if isinstance(ants_kwargs['moving_mask'], str) or isinstance(ants_kwargs['moving_mask'], Path):
                ants_kwargs['moving_mask'] = ants.from_numpy(np.transpose(imread(ants_kwargs['moving_mask']).astype(float), axes=[2,1,0]), 
                                    spacing=spacing_xyz)
            elif isinstance(ants_kwargs['moving_mask'], np.ndarray):
                ants_kwargs['moving_mask'] = ants.from_numpy(np.transpose(ants_kwargs['moving_mask'].astype(float), axes=[2,1,0]), 
                                    spacing=spacing_xyz)
        
        # load in batches and process volume by volume
        chunks = self.experiment.batch_volumes(batch_size, volumes= volumes, full_only=True)
        # create a directory for the transforms
        os.makedirs(f'{save_dir}/transforms', exist_ok=True)

        tot_volumes = 0
        for ichunk, chunk in enumerate(chunks):
            # register the chunk
            outprefix = f'{save_dir}/transforms/volume_{chunk[0]:04d}'
            corrected = _register_batch(chunk, template, spacing_xyz, registration_type,
                                        outprefix = outprefix)
            # save 
            nt, z, y, x = corrected.shape
            imwrite(f'{save_dir}/drift_corrected_movie_{ichunk:04d}.tif',
                    corrected.astype(np.int16), shape=(nt, z, y, x),
                    metadata={'axes': 'TZYX'}, imagej=True)
            if verbose:
                tot_volumes = tot_volumes + nt
                print(f"written volumes : {tot_volumes}, out of {n_reguested_volumes} requested volumes")


    def batch_dff_sliding_window(self, save_dir, batch_size, window_size, blur_sigma=None, verbose=False):
        """
        Creates 3D dff movie from raw 3D movie using sliding window. Will only use full_volumes,
        so the number of frames in the resulting movie can be smaller than in the original.

        :param blur_sigma: If not None, will apply gaussian blur in 3D with sigma = blur_sigma.
                         Can be int - then the same sigma in all 3 directions is applied,
                         or list [sz,sy,sx] for different sigma in ZYX.
        :type blur_sigma: Union(int, list)
        :param save_dir: directory into which to save the dff movie in chunks
        :type save_dir: str
        :param batch_size: number of volumes to load at once
        :type batch_size: int
        :param window_size: the size of the sliding window for the baseline estimation in volumes
        :type window_size: int
        :param verbose: whether to print the volumes that have been processed so far on the screen.
        :type verbose: bool
        """
        # TODO : make the size & digit estimation
        # TODO : write resolution into metadata

        # will only use full volumes
        volume_list = self.experiment.list_volumes()
        volume_list = volume_list[volume_list >= 0]
        n_volumes = len(volume_list)

        # will multiply dff image by this value for better visualisation later
        SCALE = 1000

        # break volumes into chunks that will be downloaded at once
        overlap = window_size - 1
        # there are more chunks than needed, this is okay:
        # the cycle will end after the first chunk that contains the end of the experiment
        chunks = [volume_list[i:i + batch_size] for i in range(0, len(volume_list), batch_size - overlap)]

        for ich, chunk in enumerate(tqdm(chunks, disable=verbose)):

            data = self.experiment.load_volumes(chunk, verbose=False)
            if blur_sigma is not None:
                data = gaussian_filter(data, blur_sigma)
            dff_img, start_tp, end_tp = get_dff(data, window_size)
            t, z, y, x = dff_img.shape

            # special case of the last chunk --> need to output the tail as well, and then break
            if chunk[-1] == (n_volumes - 1):
                end_tp = t
            # special case of the first chunk --> need to output the head as well
            if ich == 0:
                start_tp = 0

            # make sure the output fits into the int16 range:
            min_value = dff_img[start_tp:end_tp, :, :, :].min()
            max_value = dff_img[start_tp:end_tp, :, :, :].max()
            if max_value * SCALE > 32767 or min_value * SCALE < -32768:
                warnings.warn(f"Scaled DFF values outside the int16 range for scale = {SCALE},"
                              f" min: {min_value}, max: {max_value}")
            # write image
            imwrite(f'{save_dir}/dff_movie_{ich:04d}.tif',
                    (dff_img[start_tp:end_tp, :, :, :] * SCALE).astype(np.int16), shape=(end_tp - start_tp, z, y, x),
                    metadata={'axes': 'TZYX'}, imagej=True)

            if verbose:
                print(f"written frames : {chunk[start_tp]} - {chunk[end_tp - 1]}, out of {n_volumes}")
            # exit cycle the first chunk you saw the end of the experiment
            if chunk[-1] == (n_volumes - 1):
                break

    def calculate_dff(self, save_dir, batch_size, step_size, baseline_volumes,
                      resolution_xyz = [1, 1, 1], blur_sigma=None, 
                      verbose=False, save_baseline=False):
        """
        Creates 3D dff movie from raw 3D movie by processing movie in steps. Will only use full_volumes,
        so the number of frames in the resulting movie can be smaller than in the original.

        Args:
            blur_sigma (Union(int, list)): If not None, will apply gaussian blur in 3D with sigma = blur_sigma.
                            Can be int - then the same sigma in all 3 directions is applied,
                            or list [sz,sy,sx] for different sigma in ZYX.
            save_dir (str): directory into which to save the dff movie in chunks
            batch_size (int): number of volumes to load at once, must be multiple of step_size
            step_size (int): the size of the sliding window for the baseline estimation in volumes
            baseline_volumes (List[int]): volumes inside each step to use for baseline estimation
            resolution_xyz (List[int]): resolution in xyz order in um
            verbose (bool): whether to print the volumes that have been processed so far on the screen.
        """
        # TODO : make the size & digit estimation
        # TODO : write resolution into metadata

        assert batch_size % step_size == 0, "batch_size must be multiple of step_size"

        if save_baseline:
            #create directory for baseline
            os.makedirs(f'{save_dir}/baseline', exist_ok=True)

        # will only use full volumes
        volume_list = self.experiment.list_volumes()
        volume_list = volume_list[volume_list >= 0]
        n_volumes = len(volume_list)

        # will multiply dff image by this value for better visualisation later
        SCALE = 1000

        #create chunks
        chunks = self.experiment.batch_volumes(batch_size, volumes=volume_list, full_only=True)

        for ich, chunk in enumerate(tqdm(chunks, disable=verbose)):

            data = self.experiment.load_volumes(chunk, verbose=False)
            # add 1 to all pixels to aviod division by zero
            data = data + 1

            if blur_sigma is not None:
                data = gaussian_filter(data, blur_sigma)
            dff_img, baseline_img = get_dff_in_steps(data, step_size, baseline_volumes)
            t, z, y, x = dff_img.shape

            # special case of the last chunk --> need to output the tail as well, and then break
            if chunk[-1] == (n_volumes - 1):
                end_tp = t
            # special case of the first chunk --> need to output the head as well
            if ich == 0:
                start_tp = 0

            # make sure the output fits into the int16 range:
            min_value = dff_img.min()
            max_value = dff_img.max()
            if max_value * SCALE > 32767 or min_value * SCALE < -32768:
                warnings.warn(f"Scaled DFF values outside the int16 range for scale = {SCALE},"
                              f" min: {min_value}, max: {max_value}")
            # write dff image
            imwrite(f'{save_dir}/dff_movie_{ich:04d}.tif', (dff_img * SCALE).astype(np.int16), shape=(t, z, y, x),
                    resolution=(1./resolution_xyz[0], 1./resolution_xyz[1]), metadata={'spacing': resolution_xyz[2], 'unit': 'um','axes': 'TZYX'}, 
                    imagej=True)
            if save_baseline:
                # print(f"Baseline shape{baseline_img.shape}")
                # write baseline image
                imwrite(f'{save_dir}/baseline/baseline_movie_{ich:04d}.tif', baseline_img.astype(np.int16), shape=(t, z, y, x),
                        resolution=(1./resolution_xyz[0], 1./resolution_xyz[1]), metadata={'spacing': resolution_xyz[2], 'unit': 'um','axes': 'TZYX'}, 
                        imagej=True)

            if verbose:
                print(f"written frames : {chunk[0]} - {chunk[- 1]}, out of {n_volumes}")
            # exit cycle the first chunk you saw the end of the experiment
            if chunk[-1] == (n_volumes - 1):
                break


if __name__ == "__main__":
    pass



import nbformat
from nbconvert.preprocessors import ExecutePreprocessor, CellExecutionError
from .notifications import slack_notification
import os


def run_notebook(notebook, notify=True, tag_users=None):
    try:
        with open(notebook) as f:
            nb = nbformat.read(f, as_version=4)

        # waits 3 hours for the cell to finish execution
        ep = ExecutePreprocessor(timeout=3 * 60 * 60, kernel_name="numan")
    except Exception as e:
        msg = f'Error before executing the notebook {notebook}\n'
        msg += e
        print(msg)
        if notify:
            slack_notification(msg, tag_users=tag_users)
        raise e

    success = True
    try:
        out = ep.preprocess(nb)
    except Exception as e:
        out = None
        msg = f'Error executing the notebook {notebook}\n' \
              f'{e}\n' \
              f'See notebook failed_{notebook} for the traceback.'
        print(msg)
        success = False
        if notify:
            slack_notification(msg, tag_users=tag_users)
        raise
    finally:
        if success:
            if notify:
                msg = f'Successfully executed the notebook {notebook}.\n\n'
                slack_notification(msg, tag_users=tag_users)
            with open(f"done_{notebook}", mode='w', encoding='utf-8') as f:
                nbformat.write(nb, f)
        else:
            with open(f"failed_{notebook}", mode='w', encoding='utf-8') as f:
                nbformat.write(nb, f)


def wait_for_manual_step(current_notebook, notebooks_before_manual,
                         notify=False, tag_users=None):
    try:
        do_wait = False
        if notebooks_before_manual is not None:
            if current_notebook in notebooks_before_manual:
                if notify:
                    msg = f"Stopping execution, waiting for you to complete the manual step."
                    slack_notification(msg, tag_users=tag_users)
                do_wait = True
    except Exception as e:
        msg = f'Error in numan.wait_for_manual_step executing the notebook {current_notebook}\n'
        msg += e
        print(msg)
        if notify:
            slack_notification(msg, tag_users=tag_users)
        raise e
    return do_wait


def run_notebooks(notebooks, notebooks_before_manual=None, notify=True, tag_users=None, rerun=False):
    if notify:
        msg = f"Started processing notebooks in {os.getcwd()}"
        slack_notification(msg, tag_users=tag_users)

    for nb_run in notebooks:
        if rerun:
            run_notebook(nb_run, notify=notify, tag_users=tag_users)
            if wait_for_manual_step(nb_run, notebooks_before_manual,
                                    notify=notify, tag_users=tag_users):
                break
        else:
            if os.path.exists(f"done_{nb_run}"):
                msg = f"Notebook done_{nb_run} already exists, skipping {nb_run}\n"
                print(msg)
                if notify:
                    slack_notification(msg, tag_users=tag_users)
                pass
            else:
                run_notebook(nb_run, notify=notify, tag_users=tag_users)
                if wait_for_manual_step(nb_run, notebooks_before_manual,
                                        notify=notify, tag_users=tag_users):
                    break
